from __future__ import annotations

from dataclasses import dataclass
from typing import Annotated, Protocol, TypeAlias, get_args, get_origin

import pytest

import diwire._internal.markers as markers_module
from diwire import Container
from diwire._internal.markers import (
    All,
    AllMarker,
    AsyncProvider,
    Component,
    FromContext,
    FromContextMarker,
    Injected,
    InjectedMarker,
    Maybe,
    MaybeMarker,
    Provider,
    ProviderMarker,
    component_base_key,
    is_all_annotation,
    is_async_provider_annotation,
    is_from_context_annotation,
    is_maybe_annotation,
    is_provider_annotation,
    strip_all_annotation,
    strip_from_context_annotation,
    strip_maybe_annotation,
    strip_provider_annotation,
)
from diwire._internal.providers import ProviderDependenciesExtractor


class Database(Protocol):
    def query(self) -> str: ...


PrimaryDatabaseComponent: TypeAlias = Annotated[Database, Component("primary")]
ReplicaDatabaseComponent: TypeAlias = Annotated[Database, Component("replica")]


@dataclass
class PrimaryDatabase:
    def query(self) -> str:
        return "primary"


@dataclass
class ReplicaDatabase:
    def query(self) -> str:
        return "replica"


@dataclass
class Repository:
    primary: PrimaryDatabaseComponent
    replica: ReplicaDatabaseComponent


def test_component_marker_is_value_based_and_hashable() -> None:
    marker = Component("primary")

    assert marker.value == "primary"
    assert marker == Component("primary")
    assert marker != Component("replica")

    mapping = {marker: "database"}
    assert mapping[Component("primary")] == "database"


def test_provider_dependencies_extractor_preserves_component_marker() -> None:
    def build_repository(primary: PrimaryDatabaseComponent) -> Repository:
        return Repository(primary=primary, replica=ReplicaDatabase())

    extractor = ProviderDependenciesExtractor()

    dependencies = extractor.extract_from_factory(build_repository)

    assert [dependency.parameter.name for dependency in dependencies] == ["primary"]
    assert [dependency.provides for dependency in dependencies] == [PrimaryDatabaseComponent]


def test_injected_wraps_dependency_with_injected_marker() -> None:
    dependency = Injected[Database]

    assert get_origin(dependency) is Annotated
    annotation_args = get_args(dependency)
    assert annotation_args[0] is Database
    assert isinstance(annotation_args[1], InjectedMarker)


def test_injected_preserves_component_marker_metadata_when_nested() -> None:
    dependency = Injected[PrimaryDatabaseComponent]

    assert get_origin(dependency) is Annotated
    annotation_args = get_args(dependency)
    assert annotation_args[0] is Database
    assert annotation_args[1] == Component("primary")
    assert isinstance(annotation_args[2], InjectedMarker)


def test_from_context_wraps_dependency_with_marker() -> None:
    dependency = FromContext[Database]

    assert get_origin(dependency) is Annotated
    annotation_args = get_args(dependency)
    assert annotation_args[0] is Database
    assert isinstance(annotation_args[1], FromContextMarker)


def test_from_context_preserves_component_marker_metadata_when_nested() -> None:
    dependency = FromContext[PrimaryDatabaseComponent]

    assert get_origin(dependency) is Annotated
    annotation_args = get_args(dependency)
    assert annotation_args[0] is Database
    assert annotation_args[1] == Component("primary")
    assert isinstance(annotation_args[2], FromContextMarker)


def test_from_context_helpers_detect_and_strip_marker() -> None:
    dependency = FromContext[PrimaryDatabaseComponent]

    assert is_from_context_annotation(dependency) is True
    assert strip_from_context_annotation(dependency) == PrimaryDatabaseComponent
    assert is_from_context_annotation(int) is False
    assert strip_from_context_annotation(int) is int


def test_maybe_wraps_dependency_with_marker() -> None:
    dependency = Maybe[Database]

    assert get_origin(dependency) is Annotated
    annotation_args = get_args(dependency)
    assert annotation_args[0] is Database
    assert isinstance(annotation_args[1], MaybeMarker)


def test_maybe_preserves_component_marker_metadata_when_nested() -> None:
    dependency = Maybe[PrimaryDatabaseComponent]

    assert get_origin(dependency) is Annotated
    annotation_args = get_args(dependency)
    assert annotation_args[0] is Database
    assert annotation_args[1] == Component("primary")
    assert isinstance(annotation_args[2], MaybeMarker)


def test_maybe_helpers_detect_and_strip_marker() -> None:
    dependency = Maybe[PrimaryDatabaseComponent]

    assert is_maybe_annotation(dependency) is True
    assert strip_maybe_annotation(dependency) == PrimaryDatabaseComponent
    assert is_maybe_annotation(int) is False
    assert strip_maybe_annotation(int) is int


def test_provider_wraps_dependency_with_provider_marker() -> None:
    dependency = Provider[Database]

    assert get_origin(dependency) is Annotated
    annotation_args = get_args(dependency)
    assert annotation_args[0] is Database
    assert isinstance(annotation_args[1], ProviderMarker)
    assert annotation_args[1].dependency_key is Database
    assert annotation_args[1].is_async is False


def test_async_provider_wraps_dependency_with_async_provider_marker() -> None:
    dependency = AsyncProvider[Database]

    assert get_origin(dependency) is Annotated
    annotation_args = get_args(dependency)
    assert annotation_args[0] is Database
    assert isinstance(annotation_args[1], ProviderMarker)
    assert annotation_args[1].dependency_key is Database
    assert annotation_args[1].is_async is True


def test_provider_preserves_component_marker_metadata_when_nested() -> None:
    dependency = Provider[PrimaryDatabaseComponent]

    assert get_origin(dependency) is Annotated
    annotation_args = get_args(dependency)
    assert annotation_args[0] is Database
    assert annotation_args[1] == Component("primary")
    assert isinstance(annotation_args[2], ProviderMarker)
    assert annotation_args[2].dependency_key == PrimaryDatabaseComponent


def test_provider_helpers_detect_strip_and_check_async_flags() -> None:
    sync_dependency = Provider[PrimaryDatabaseComponent]
    async_dependency = AsyncProvider[PrimaryDatabaseComponent]

    assert is_provider_annotation(sync_dependency) is True
    assert is_provider_annotation(async_dependency) is True
    assert strip_provider_annotation(sync_dependency) == PrimaryDatabaseComponent
    assert strip_provider_annotation(async_dependency) == PrimaryDatabaseComponent
    assert strip_provider_annotation(int) is int
    assert is_async_provider_annotation(sync_dependency) is False
    assert is_async_provider_annotation(async_dependency) is True
    assert is_async_provider_annotation(int) is False


def test_all_wraps_dependency_with_all_marker() -> None:
    dependency = All[Database]

    assert get_origin(dependency) is Annotated
    annotation_args = get_args(dependency)
    assert annotation_args[0] is Database
    assert isinstance(annotation_args[1], AllMarker)
    assert annotation_args[1].dependency_key is Database


def test_all_strips_annotated_item_to_base_key() -> None:
    dependency = All[PrimaryDatabaseComponent]

    assert get_origin(dependency) is Annotated
    annotation_args = get_args(dependency)
    assert annotation_args[0] is Database
    assert len(annotation_args) == 2
    assert isinstance(annotation_args[1], AllMarker)
    assert annotation_args[1].dependency_key is Database


def test_injected_all_preserves_all_marker_metadata_when_nested() -> None:
    dependency = Injected[All[Database]]

    assert get_origin(dependency) is Annotated
    annotation_args = get_args(dependency)
    assert annotation_args[0] is Database
    assert isinstance(annotation_args[1], AllMarker)
    assert isinstance(annotation_args[2], InjectedMarker)


def test_all_helpers_detect_and_strip_marker() -> None:
    dependency = All[Database]

    assert is_all_annotation(dependency) is True
    assert strip_all_annotation(dependency) is Database
    assert is_all_annotation(int) is False
    assert strip_all_annotation(int) is int


def test_component_base_key_detects_component_keys() -> None:
    unrelated_annotated = Annotated[Database, "unrelated"]

    assert component_base_key(PrimaryDatabaseComponent) is Database
    assert component_base_key(Database) is None
    assert component_base_key(Injected[PrimaryDatabaseComponent]) is Database
    assert component_base_key(unrelated_annotated) is None


def test_injected_provider_preserves_provider_metadata() -> None:
    dependency = Injected[Provider[PrimaryDatabaseComponent]]

    assert get_origin(dependency) is Annotated
    annotation_args = get_args(dependency)
    assert annotation_args[0] is Database
    assert annotation_args[1] == Component("primary")
    assert isinstance(annotation_args[2], ProviderMarker)
    assert annotation_args[2].dependency_key == PrimaryDatabaseComponent
    assert isinstance(annotation_args[3], InjectedMarker)


def test_is_from_context_annotation_handles_invalid_annotated_shape(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setattr(markers_module, "get_origin", lambda _annotation: Annotated)
    monkeypatch.setattr(markers_module, "get_args", lambda _annotation: (int,))

    sentinel = object()

    assert markers_module.is_from_context_annotation(sentinel) is False
    assert markers_module.is_maybe_annotation(sentinel) is False
    assert markers_module.is_provider_annotation(sentinel) is False
    assert markers_module.is_all_annotation(sentinel) is False
    assert markers_module.strip_all_annotation(sentinel) is sentinel
    assert markers_module.strip_maybe_annotation(sentinel) is sentinel
    assert markers_module.component_base_key(sentinel) is None


def test_provider_dependencies_extractor_preserves_injected_component_dependency() -> None:
    def build_repository(primary: Injected[PrimaryDatabaseComponent]) -> Repository:
        return Repository(primary=primary, replica=ReplicaDatabase())

    extractor = ProviderDependenciesExtractor()

    dependencies = extractor.extract_from_factory(build_repository)

    assert [dependency.parameter.name for dependency in dependencies] == ["primary"]
    annotation_args = get_args(dependencies[0].provides)
    assert len(annotation_args) == 3
    assert annotation_args[0] is Database
    assert annotation_args[1] == Component("primary")
    assert isinstance(annotation_args[2], InjectedMarker)


def test_container_resolves_distinct_component_registrations() -> None:
    def build_primary() -> PrimaryDatabaseComponent:
        return PrimaryDatabase()

    def build_replica() -> ReplicaDatabaseComponent:
        return ReplicaDatabase()

    container = Container()
    container.add_factory(build_primary)
    container.add_factory(build_replica)

    primary = container.resolve(PrimaryDatabaseComponent)
    replica = container.resolve(ReplicaDatabaseComponent)

    assert isinstance(primary, PrimaryDatabase)
    assert isinstance(replica, ReplicaDatabase)
    assert primary.query() == "primary"
    assert replica.query() == "replica"


def test_container_injects_component_marked_dependencies() -> None:
    def build_primary() -> PrimaryDatabaseComponent:
        return PrimaryDatabase()

    def build_replica() -> ReplicaDatabaseComponent:
        return ReplicaDatabase()

    container = Container()
    container.add_factory(build_primary)
    container.add_factory(build_replica)
    container.add(Repository)

    repository = container.resolve(Repository)

    assert isinstance(repository.primary, PrimaryDatabase)
    assert isinstance(repository.replica, ReplicaDatabase)



"""Focused example: ``add_factory`` for custom build logic."""

from __future__ import annotations

from dataclasses import dataclass

from diwire import Container, Lifetime


@dataclass(slots=True)
class Service:
    value: str


def main() -> None:
    container = Container()
    build_state = {"count": 0}

    def build_service() -> Service:
        build_state["count"] += 1
        return Service(value=f"built-{build_state['count']}")

    container.add_factory(
        build_service,
        provides=Service,
        lifetime=Lifetime.TRANSIENT,
    )

    first = container.resolve(Service)
    second = container.resolve(Service)
    print(f"factory_custom_logic={first.value}")  # => factory_custom_logic=built-1
    print(f"factory_is_transient={first is not second}")  # => factory_is_transient=True


if __name__ == "__main__":
    main()



from __future__ import annotations

import pytest

from diwire._internal.resolvers.assembly.assembly_engine import Environment


def test_variable_interpolation_renders_context_value() -> None:
    assembly = Environment().from_string("hello {{ name }}")

    rendered = assembly.render(name="world")

    assert rendered == "hello world"


def test_if_condition_uses_truthy_branch() -> None:
    assembly = Environment().from_string("{% if is_enabled %}yes{% endif %}")

    rendered = assembly.render(is_enabled=True)

    assert rendered == "yes"


def test_if_condition_uses_falsy_branch() -> None:
    assembly = Environment().from_string("{% if is_enabled %}yes{% else %}no{% endif %}")

    rendered = assembly.render(is_enabled=False)

    assert rendered == "no"


def test_if_equality_condition_matches_literal() -> None:
    assembly = Environment().from_string('{% if mode == "async" %}ok{% else %}skip{% endif %}')

    rendered = assembly.render(mode="async")

    assert rendered == "ok"


def test_if_equality_condition_does_not_match_literal() -> None:
    assembly = Environment().from_string('{% if mode == "async" %}ok{% else %}skip{% endif %}')

    rendered = assembly.render(mode="thread")

    assert rendered == "skip"


def test_nested_if_blocks_render_expected_branch() -> None:
    assembly = Environment().from_string(
        "{% if outer %}A{% if inner %}B{% else %}C{% endif %}{% else %}D{% endif %}",
    )

    rendered = assembly.render(outer=True, inner=False)

    assert rendered == "AC"


def test_environment_rejects_autoescape_true() -> None:
    with pytest.raises(ValueError, match="autoescape=True"):
        Environment(autoescape=True)


def test_parser_rejects_unknown_block_tag() -> None:
    with pytest.raises(ValueError, match="Unsupported assembly tag"):
        Environment().from_string("{% for item in items %}{{ item }}{% endfor %}")


def test_parser_rejects_unclosed_if_block() -> None:
    with pytest.raises(ValueError, match="missing endif"):
        Environment().from_string("{% if condition %}x")


def test_parser_rejects_unclosed_if_block_with_else_branch() -> None:
    with pytest.raises(ValueError, match="missing endif"):
        Environment().from_string("{% if condition %}x{% else %}y")


def test_parser_rejects_unexpected_else_block() -> None:
    with pytest.raises(ValueError, match="Unexpected block tag 'else'"):
        Environment().from_string("{% else %}")


def test_parser_rejects_unclosed_variable_tag() -> None:
    with pytest.raises(ValueError, match="Unclosed variable tag"):
        Environment().from_string("{{ value")


def test_render_raises_for_missing_variable() -> None:
    assembly = Environment().from_string("{{ value }}")

    with pytest.raises(ValueError, match="Missing assembly variable 'value'"):
        assembly.render()


def test_render_raises_for_missing_if_condition_variable() -> None:
    assembly = Environment().from_string("{% if is_enabled %}yes{% endif %}")

    with pytest.raises(ValueError, match="Missing assembly variable 'is_enabled'"):
        assembly.render()


def test_parser_rejects_unsupported_variable_expression() -> None:
    with pytest.raises(ValueError, match="Unsupported variable expression"):
        Environment().from_string("{{ value.name }}")


def test_parser_rejects_unsupported_if_condition_expression() -> None:
    with pytest.raises(ValueError, match="Unsupported if condition"):
        Environment().from_string('{% if mode != "async" %}x{% endif %}')



"""Open-generics constraints deep dive.

This focused script shows constrained ``TypeVar`` behavior with both valid and
invalid resolutions.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Generic, TypeVar, cast

from diwire import Container
from diwire.exceptions import DIWireInvalidGenericTypeArgumentError

Allowed = TypeVar("Allowed", int, str)


class ConstrainedBox(Generic[Allowed]):
    pass


@dataclass(slots=True)
class ConstrainedBoxImpl(ConstrainedBox[Allowed]):
    type_arg: type[Allowed]


def main() -> None:
    container = Container()
    container.add(ConstrainedBoxImpl, provides=ConstrainedBox)

    valid_int = container.resolve(ConstrainedBox[int])
    valid_str = container.resolve(ConstrainedBox[str])

    if cast("ConstrainedBoxImpl[int]", valid_int).type_arg is not int:
        msg = "Expected int constrained type argument"
        raise TypeError(msg)
    if cast("ConstrainedBoxImpl[str]", valid_str).type_arg is not str:
        msg = "Expected str constrained type argument"
        raise TypeError(msg)

    invalid_key = cast("Any", ConstrainedBox)[float]
    try:
        container.resolve(invalid_key)
    except DIWireInvalidGenericTypeArgumentError:
        return

    msg = "Expected DIWireInvalidGenericTypeArgumentError for constrained TypeVar"
    raise TypeError(msg)


if __name__ == "__main__":
    main()



"""Focused example: scoped resource cleanup on scope exit."""

from __future__ import annotations

from collections.abc import Generator

from diwire import Container, Lifetime, Scope


class ScopedResource:
    pass


def main() -> None:
    container = Container()
    state = {"closed": 0}

    def provide_resource() -> Generator[ScopedResource, None, None]:
        try:
            yield ScopedResource()
        finally:
            state["closed"] += 1

    container.add_generator(
        provide_resource,
        provides=ScopedResource,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope() as request_scope:
        _ = request_scope.resolve(ScopedResource)
        closed_inside_scope = state["closed"]

    closed_after_exit = state["closed"]
    print(
        f"scoped_cleanup_after_exit={closed_inside_scope == 0 and closed_after_exit == 1}",
    )  # => scoped_cleanup_after_exit=True


if __name__ == "__main__":
    main()



from __future__ import annotations

import inspect

from diwire.exceptions import DIWireInvalidRegistrationError


class DependecyRegistrationValidator:
    """Validates dependency registrations before creating provider specs."""

    def validate_concrete_type(self, concrete_type: object) -> None:
        """Validate that a concrete provider is instantiable.

        Args:
            concrete_type: Concrete class provider to inspect or validate.

        """
        if not inspect.isclass(concrete_type):
            msg = f"Concrete provider must be a class, got {concrete_type!r}."
            raise DIWireInvalidRegistrationError(msg)

        if inspect.isabstract(concrete_type):
            msg = f"Concrete provider '{concrete_type.__qualname__}' cannot be an abstract class."
            raise DIWireInvalidRegistrationError(msg)



from __future__ import annotations

from typing import Any

import punq
import rodi
from dishka import Provider
from wireup import injectable

from diwire import Lifetime
from tests.benchmarks.dishka_helpers import DishkaBenchmarkScope, make_dishka_benchmark_container
from tests.benchmarks.helpers import make_diwire_benchmark_container, run_benchmark
from tests.benchmarks.wireup_helpers import make_wireup_benchmark_container


@injectable(lifetime="transient")
class _DepA:
    pass


@injectable(lifetime="transient")
class _DepB:
    pass


@injectable(lifetime="transient")
class _DepC:
    pass


@injectable(lifetime="transient")
class _DepD:
    pass


@injectable(lifetime="transient")
class _DepE:
    pass


@injectable(lifetime="transient")
class _Root:
    def __init__(
        self,
        dep_a: _DepA,
        dep_b: _DepB,
        dep_c: _DepC,
        dep_d: _DepD,
        dep_e: _DepE,
    ) -> None:
        self.dep_a = dep_a
        self.dep_b = dep_b
        self.dep_c = dep_c
        self.dep_d = dep_d
        self.dep_e = dep_e


def test_benchmark_diwire_resolve_wide_transient_graph(benchmark: Any) -> None:
    container = make_diwire_benchmark_container()
    container.add(_DepA, lifetime=Lifetime.TRANSIENT)
    container.add(_DepB, lifetime=Lifetime.TRANSIENT)
    container.add(_DepC, lifetime=Lifetime.TRANSIENT)
    container.add(_DepD, lifetime=Lifetime.TRANSIENT)
    container.add(_DepE, lifetime=Lifetime.TRANSIENT)
    container.add(_Root, lifetime=Lifetime.TRANSIENT)
    container.compile()
    first = container.resolve(_Root)
    second = container.resolve(_Root)
    assert first is not second
    assert first.dep_a is not second.dep_a
    assert first.dep_b is not second.dep_b
    assert first.dep_c is not second.dep_c
    assert first.dep_d is not second.dep_d
    assert first.dep_e is not second.dep_e

    def bench_diwire_wide_graph() -> None:
        _ = container.resolve(_Root)

    run_benchmark(benchmark, bench_diwire_wide_graph, iterations=25_000)


def test_benchmark_rodi_resolve_wide_transient_graph(benchmark: Any) -> None:
    rodi_container = rodi.Container()
    rodi_container.add_transient(_DepA)
    rodi_container.add_transient(_DepB)
    rodi_container.add_transient(_DepC)
    rodi_container.add_transient(_DepD)
    rodi_container.add_transient(_DepE)
    rodi_container.add_transient(_Root)
    services = rodi_container.build_provider()
    first = services.get(_Root)
    second = services.get(_Root)
    assert first is not second
    assert first.dep_a is not second.dep_a
    assert first.dep_b is not second.dep_b
    assert first.dep_c is not second.dep_c
    assert first.dep_d is not second.dep_d
    assert first.dep_e is not second.dep_e

    def bench_rodi_wide_graph() -> None:
        _ = services.get(_Root)

    run_benchmark(benchmark, bench_rodi_wide_graph, iterations=25_000)


def test_benchmark_dishka_resolve_wide_transient_graph(benchmark: Any) -> None:
    provider = Provider(scope=DishkaBenchmarkScope.APP)
    provider.provide(_DepA, scope=DishkaBenchmarkScope.APP, cache=False)
    provider.provide(_DepB, scope=DishkaBenchmarkScope.APP, cache=False)
    provider.provide(_DepC, scope=DishkaBenchmarkScope.APP, cache=False)
    provider.provide(_DepD, scope=DishkaBenchmarkScope.APP, cache=False)
    provider.provide(_DepE, scope=DishkaBenchmarkScope.APP, cache=False)
    provider.provide(_Root, scope=DishkaBenchmarkScope.APP, cache=False)
    container = make_dishka_benchmark_container(provider)
    first = container.get(_Root)
    second = container.get(_Root)
    assert first is not second
    assert first.dep_a is not second.dep_a
    assert first.dep_b is not second.dep_b
    assert first.dep_c is not second.dep_c
    assert first.dep_d is not second.dep_d
    assert first.dep_e is not second.dep_e

    def bench_dishka_wide_graph() -> None:
        _ = container.get(_Root)

    run_benchmark(benchmark, bench_dishka_wide_graph, iterations=25_000)


def test_benchmark_wireup_resolve_wide_transient_graph(benchmark: Any) -> None:
    container = make_wireup_benchmark_container(_DepA, _DepB, _DepC, _DepD, _DepE, _Root)
    with container.enter_scope() as scope:
        first = scope.get(_Root)
        second = scope.get(_Root)
        assert first is not second
        assert first.dep_a is not second.dep_a
        assert first.dep_b is not second.dep_b
        assert first.dep_c is not second.dep_c
        assert first.dep_d is not second.dep_d
        assert first.dep_e is not second.dep_e

        def bench_wireup_wide_graph() -> None:
            _ = scope.get(_Root)

        run_benchmark(benchmark, bench_wireup_wide_graph, iterations=25_000)


def test_benchmark_punq_resolve_wide_transient_graph(benchmark: Any) -> None:
    container = punq.Container()
    container.register(_DepA)
    container.register(_DepB)
    container.register(_DepC)
    container.register(_DepD)
    container.register(_DepE)
    container.register(_Root)
    first = container.resolve(_Root)
    second = container.resolve(_Root)
    assert first is not second
    assert first.dep_a is not second.dep_a
    assert first.dep_b is not second.dep_b
    assert first.dep_c is not second.dep_c
    assert first.dep_d is not second.dep_d
    assert first.dep_e is not second.dep_e

    def bench_punq_wide_graph() -> None:
        _ = container.resolve(_Root)

    run_benchmark(benchmark, bench_punq_wide_graph, iterations=25_000)



"""Focused example: async-generator provider cleanup on async scope exit."""

from __future__ import annotations

import asyncio
from collections.abc import AsyncGenerator

from diwire import Container, Lifetime, Scope


class AsyncResource:
    pass


async def main() -> None:
    container = Container()
    state = {"closed": 0}

    async def provide_async_resource() -> AsyncGenerator[AsyncResource, None]:
        try:
            yield AsyncResource()
        finally:
            await asyncio.sleep(0)
            state["closed"] += 1

    container.add_generator(
        provide_async_resource,
        provides=AsyncResource,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    async with container.enter_scope() as request_scope:
        _ = await request_scope.aresolve(AsyncResource)
        closed_before_exit = state["closed"]

    print(
        f"async_cleanup_after_exit={closed_before_exit == 0 and state['closed'] == 1}",
    )  # => async_cleanup_after_exit=True


if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

import asyncio
import inspect
from collections.abc import Callable
from typing import Any, cast

from diwire import Container, Injected
from diwire.integrations.pytest_plugin import pytest_pycollect_makeitem, pytest_pyfunc_call


class _Service:
    def __init__(self, value: str = "service") -> None:
        self.value = value


class _DummyCollector:
    def __init__(self, *, is_test_function: bool) -> None:
        self._is_test_function = is_test_function

    def istestfunction(self, obj: object, name: str) -> bool:
        _ = obj, name
        return self._is_test_function


class _DummyPyFuncItem:
    def __init__(self, *, obj: Callable[..., Any], container: Container | None) -> None:
        self.obj = obj
        if container is not None:
            self._diwire_container = container


def test_pycollect_makeitem_ignores_non_callable_objects() -> None:
    collector = _DummyCollector(is_test_function=True)

    result = pytest_pycollect_makeitem(collector=collector, name="test_value", obj=1)

    assert result is None


def test_pycollect_makeitem_ignores_non_test_callables() -> None:
    collector = _DummyCollector(is_test_function=False)

    def helper(value: int, service: Injected[_Service]) -> None:
        _ = value, service

    original_signature = inspect.signature(helper)
    result = pytest_pycollect_makeitem(collector=collector, name="helper", obj=helper)

    assert result is None
    assert inspect.signature(helper) == original_signature


def test_pycollect_makeitem_rewrites_signature_for_injected_parameters() -> None:
    collector = _DummyCollector(is_test_function=True)

    def test_handler(value: int, service: Injected[_Service]) -> tuple[int, _Service]:
        return value, service

    assert tuple(inspect.signature(test_handler).parameters) == ("value", "service")
    result = pytest_pycollect_makeitem(
        collector=collector,
        name="test_handler",
        obj=test_handler,
    )

    assert result is None
    assert tuple(inspect.signature(test_handler).parameters) == ("value",)


def test_pyfunc_call_passes_through_when_no_injected_parameters() -> None:
    def test_handler(value: int) -> int:
        return value

    item = _DummyPyFuncItem(obj=test_handler, container=Container())
    hook = pytest_pyfunc_call(cast("Any", item))
    next(hook)

    assert item.obj is test_handler
    next(hook, None)
    assert item.obj is test_handler


def test_pyfunc_call_wraps_injected_callable_and_restores_original() -> None:
    container = Container()
    default_dependency = _Service("container")
    container.add_instance(default_dependency, provides=_Service)

    def test_handler(service: Injected[_Service]) -> _Service:
        return service

    item = _DummyPyFuncItem(obj=test_handler, container=container)
    hook = pytest_pyfunc_call(cast("Any", item))
    next(hook)

    wrapped = cast("Callable[..., _Service]", item.obj)
    override_dependency = _Service("override")
    assert wrapped() is default_dependency
    assert wrapped(service=override_dependency) is override_dependency

    next(hook, None)
    assert item.obj is test_handler


def test_pyfunc_call_passes_through_when_container_state_is_missing() -> None:
    def test_handler(service: Injected[_Service]) -> _Service:
        return service

    item = _DummyPyFuncItem(obj=test_handler, container=None)
    hook = pytest_pyfunc_call(cast("Any", item))
    next(hook)

    assert item.obj is test_handler
    next(hook, None)
    assert item.obj is test_handler


def test_pyfunc_call_wraps_async_injected_callable_and_restores_original() -> None:
    container = Container()
    default_dependency = _Service("container")
    container.add_instance(default_dependency, provides=_Service)

    async def test_handler(service: Injected[_Service]) -> _Service:
        return service

    item = _DummyPyFuncItem(obj=test_handler, container=container)
    hook = pytest_pyfunc_call(cast("Any", item))
    next(hook)

    assert asyncio.run(item.obj()) is default_dependency

    next(hook, None)
    assert item.obj is test_handler



from __future__ import annotations

import importlib.util
import sys
from pathlib import Path
from types import ModuleType
from typing import Any


def _find_repo_root(start: Path) -> Path:
    for candidate in (start, *start.parents):
        if (candidate / "pyproject.toml").is_file() and (candidate / "src" / "diwire").is_dir():
            return candidate
    msg = f"Could not locate repository root from {start}"
    raise AssertionError(msg)


def _load_sitemap_module(repo_root: Path) -> ModuleType:
    module_path = repo_root / "docs" / "_extensions" / "diwire_sitemap.py"
    module_name = "diwire_sitemap_test_module"
    spec = importlib.util.spec_from_file_location(module_name, module_path)
    if spec is None or spec.loader is None:
        msg = f"Could not load module from {module_path}"
        raise AssertionError(msg)

    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module


def _write_html(path: Path) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text("<html><body>ok</body></html>\n", encoding="utf-8")


REPO_ROOT = _find_repo_root(Path(__file__).resolve())
SITEMAP: Any = _load_sitemap_module(REPO_ROOT)


def test_canonical_path_from_html_rel_maps_expected_shapes() -> None:
    assert SITEMAP._canonical_path_from_html_rel("index.html") == "/"
    assert SITEMAP._canonical_path_from_html_rel("core/index.html") == "/core/"
    assert SITEMAP._canonical_path_from_html_rel("core/async.html") == "/core/async/"


def test_write_extensionless_redirects_creates_redirect_for_non_index_pages(tmp_path: Path) -> None:
    source_html = tmp_path / "core" / "async.html"
    _write_html(source_html)

    SITEMAP._write_extensionless_redirects(
        tmp_path,
        html_files=[source_html],
        baseurl="https://docs.example.dev",
    )

    redirect_path = tmp_path / "core" / "async" / "index.html"
    assert redirect_path.is_file()

    redirect_html = redirect_path.read_text(encoding="utf-8")
    assert '<link rel="canonical" href="https://docs.example.dev/core/async/">' in redirect_html
    assert '<meta http-equiv="refresh" content="0; url=../async.html">' in redirect_html
    assert 'window.location.replace("../async.html");' in redirect_html


def test_write_extensionless_redirects_skips_index_pages(tmp_path: Path) -> None:
    root_index = tmp_path / "index.html"
    section_index = tmp_path / "core" / "index.html"
    _write_html(root_index)
    _write_html(section_index)

    SITEMAP._write_extensionless_redirects(
        tmp_path,
        html_files=[root_index, section_index],
        baseurl="https://docs.example.dev",
    )

    assert not (tmp_path / "index" / "index.html").exists()
    assert not (tmp_path / "core" / "index" / "index.html").exists()


def test_build_entries_uses_extensionless_canonical_urls(tmp_path: Path) -> None:
    _write_html(tmp_path / "index.html")
    _write_html(tmp_path / "core" / "index.html")
    _write_html(tmp_path / "core" / "async.html")

    entries = SITEMAP._build_entries(tmp_path, baseurl="https://docs.example.dev")
    locs = {entry.loc for entry in entries}

    assert locs == {
        "https://docs.example.dev/",
        "https://docs.example.dev/core/",
        "https://docs.example.dev/core/async/",
    }


def test_build_entries_ignores_generated_redirect_index_pages(tmp_path: Path) -> None:
    _write_html(tmp_path / "core" / "async.html")
    _write_html(tmp_path / "core" / "async" / "index.html")

    entries = SITEMAP._build_entries(tmp_path, baseurl="https://docs.example.dev")
    locs = [entry.loc for entry in entries]

    assert locs == ["https://docs.example.dev/core/async/"]



"""Resolve all implementations with ``All[T]`` (base + components).

This module demonstrates how to collect a plugin stack by combining:

- the plain registration for a base type ``T`` (if present), and
- all component registrations keyed as ``Annotated[T, Component(...)]``.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Annotated, Protocol, TypeAlias

from diwire import All, Component, Container, Injected, resolver_context


class EventHandler(Protocol):
    def handle(self, event: str) -> str: ...


@dataclass(frozen=True, slots=True)
class BaseHandler:
    def handle(self, event: str) -> str:
        return f"base:{event}"


@dataclass(frozen=True, slots=True)
class LoggingHandler:
    def handle(self, event: str) -> str:
        return f"logging:{event}"


@dataclass(frozen=True, slots=True)
class MetricsHandler:
    def handle(self, event: str) -> str:
        return f"metrics:{event}"


Logging: TypeAlias = Annotated[EventHandler, Component("logging")]
Metrics: TypeAlias = Annotated[EventHandler, Component("metrics")]


def main() -> None:
    container = Container()

    container.add(BaseHandler, provides=EventHandler)
    container.add(LoggingHandler, provides=Logging)
    container.add(MetricsHandler, provides=Metrics)

    handlers = container.resolve(All[EventHandler])
    print(
        [handler.handle("evt") for handler in handlers],
    )  # => ['base:evt', 'logging:evt', 'metrics:evt']

    @resolver_context.inject
    def dispatch(event: str, handlers: Injected[All[EventHandler]]) -> tuple[str, ...]:
        return tuple(handler.handle(event) for handler in handlers)

    print(dispatch("evt"))  # => ('base:evt', 'logging:evt', 'metrics:evt')


if __name__ == "__main__":
    main()



from __future__ import annotations

import types
from typing import Any, TypeGuard


def is_runtime_class(candidate: object) -> TypeGuard[type[Any]]:
    """Return true when candidate is a runtime class safe for class-only operations.

    Args:
        candidate: Value being checked for eligibility or runtime type constraints.

    """
    return isinstance(candidate, type) and not isinstance(candidate, types.GenericAlias)


__all__ = ["is_runtime_class"]



from __future__ import annotations

import re
from dataclasses import dataclass

_IDENTIFIER_PATTERN = re.compile(r"^[A-Za-z_][A-Za-z0-9_]*$")
_EQUALITY_PATTERN = re.compile(
    r'^(?P<identifier>[A-Za-z_][A-Za-z0-9_]*)\s*==\s*"(?P<literal>[^"]*)"$',
)


@dataclass(frozen=True, slots=True)
class _TextToken:
    value: str


@dataclass(frozen=True, slots=True)
class _VariableToken:
    expression: str


@dataclass(frozen=True, slots=True)
class _BlockToken:
    expression: str


_Token = _TextToken | _VariableToken | _BlockToken


@dataclass(frozen=True, slots=True)
class _Condition:
    identifier: str
    expected_literal: str | None


@dataclass(frozen=True, slots=True)
class _TextNode:
    value: str


@dataclass(frozen=True, slots=True)
class _VariableNode:
    identifier: str


@dataclass(frozen=True, slots=True)
class _IfNode:
    condition: _Condition
    truthy_nodes: tuple[_Node, ...]
    falsy_nodes: tuple[_Node, ...]


_Node = _TextNode | _VariableNode | _IfNode


class Environment:
    """Minimal environment with a lightweight block-based interface used by resolver generation."""

    def __init__(self, *, autoescape: bool = False) -> None:
        if autoescape:
            msg = "autoescape=True is not supported by diwire assembly engine."
            raise ValueError(msg)

    def from_string(self, text: str) -> AssemblySnippet:
        """Compile an assembly string into a renderable assembly snippet object.

        Args:
            text: Assembly source text to compile.

        """
        parser = _Parser(tokens=_tokenize(text))
        return AssemblySnippet(nodes=parser.parse())


class AssemblySnippet:
    """Compiled mini assembly snippet."""

    def __init__(self, *, nodes: tuple[_Node, ...]) -> None:
        self._nodes = nodes

    def render(self, **context: object) -> str:
        """Render an assembly snippet with keyword-only context variables.

        Args:
            context: Optional mapping of context values to bind in the entered scope.

        """
        return _render_nodes(nodes=self._nodes, context=context)


class _Parser:
    def __init__(self, *, tokens: tuple[_Token, ...]) -> None:
        self._tokens = tokens
        self._position = 0

    def parse(self) -> tuple[_Node, ...]:
        nodes, stop_tag = self._parse_nodes(stop_tags=frozenset())
        if stop_tag is not None:
            msg = f"Unexpected block tag '{stop_tag}'."
            raise ValueError(msg)
        return tuple(nodes)

    def _parse_nodes(self, *, stop_tags: frozenset[str]) -> tuple[list[_Node], str | None]:
        nodes: list[_Node] = []

        while self._position < len(self._tokens):
            token = self._tokens[self._position]

            if isinstance(token, _TextToken):
                nodes.append(_TextNode(value=token.value))
                self._position += 1
                continue

            if isinstance(token, _VariableToken):
                nodes.append(
                    _VariableNode(
                        identifier=_parse_identifier(
                            expression=token.expression,
                            expression_kind="variable",
                        ),
                    ),
                )
                self._position += 1
                continue

            tag = token.expression
            if tag in stop_tags:
                self._position += 1
                return nodes, tag

            if tag in {"else", "endif"}:
                msg = f"Unexpected block tag '{tag}'."
                raise ValueError(msg)

            if tag.startswith("if "):
                self._position += 1
                condition = _parse_condition(expression=tag[3:].strip())
                truthy_nodes, stop_tag = self._parse_nodes(
                    stop_tags=frozenset({"else", "endif"}),
                )
                if stop_tag is None:
                    msg = "Unclosed if block: missing endif."
                    raise ValueError(msg)

                falsy_nodes: tuple[_Node, ...] = ()
                if stop_tag == "else":
                    branch_nodes, else_stop = self._parse_nodes(stop_tags=frozenset({"endif"}))
                    if else_stop != "endif":
                        msg = "Unclosed if block: missing endif."
                        raise ValueError(msg)
                    falsy_nodes = tuple(branch_nodes)

                nodes.append(
                    _IfNode(
                        condition=condition,
                        truthy_nodes=tuple(truthy_nodes),
                        falsy_nodes=falsy_nodes,
                    ),
                )
                continue

            msg = f"Unsupported assembly tag '{tag}'."
            raise ValueError(msg)

        return nodes, None


def _tokenize(text: str) -> tuple[_Token, ...]:
    tokens: list[_Token] = []
    cursor = 0
    text_length = len(text)

    while cursor < text_length:
        next_tag = _find_next_tag(text=text, cursor=cursor)
        if next_tag is None:
            tokens.append(_TextToken(value=text[cursor:]))
            break

        tag_start, tag_kind = next_tag
        if tag_start > cursor:
            tokens.append(_TextToken(value=text[cursor:tag_start]))

        if tag_kind == "variable":
            end = text.find("}}", tag_start + 2)
            if end == -1:
                msg = "Unclosed variable tag."
                raise ValueError(msg)
            expression = text[tag_start + 2 : end].strip()
            if not expression:
                msg = "Variable tag cannot be empty."
                raise ValueError(msg)
            tokens.append(_VariableToken(expression=expression))
            cursor = end + 2
            continue

        end = text.find("%}", tag_start + 2)
        if end == -1:
            msg = "Unclosed block tag."
            raise ValueError(msg)
        expression = text[tag_start + 2 : end].strip()
        if not expression:
            msg = "Block tag cannot be empty."
            raise ValueError(msg)
        tokens.append(_BlockToken(expression=expression))
        cursor = end + 2

    return tuple(tokens)


def _find_next_tag(*, text: str, cursor: int) -> tuple[int, str] | None:
    variable_start = text.find("{{", cursor)
    block_start = text.find("{%", cursor)

    if variable_start == -1 and block_start == -1:
        return None
    if variable_start == -1:
        return block_start, "block"
    if block_start == -1:
        return variable_start, "variable"
    if variable_start < block_start:
        return variable_start, "variable"
    return block_start, "block"


def _parse_identifier(*, expression: str, expression_kind: str) -> str:
    if _IDENTIFIER_PATTERN.fullmatch(expression):
        return expression
    msg = f"Unsupported {expression_kind} expression '{expression}'."
    raise ValueError(msg)


def _parse_condition(*, expression: str) -> _Condition:
    equality_match = _EQUALITY_PATTERN.fullmatch(expression)
    if equality_match is not None:
        return _Condition(
            identifier=equality_match.group("identifier"),
            expected_literal=equality_match.group("literal"),
        )

    if _IDENTIFIER_PATTERN.fullmatch(expression):
        return _Condition(identifier=expression, expected_literal=None)

    msg = f"Unsupported if condition '{expression}'."
    raise ValueError(msg)


def _render_nodes(*, nodes: tuple[_Node, ...], context: dict[str, object]) -> str:
    rendered_parts: list[str] = []

    for node in nodes:
        if isinstance(node, _TextNode):
            rendered_parts.append(node.value)
            continue

        if isinstance(node, _VariableNode):
            rendered_parts.append(
                str(_resolve_context_value(context=context, identifier=node.identifier)),
            )
            continue

        condition_matched = _evaluate_condition(condition=node.condition, context=context)
        branch = node.truthy_nodes if condition_matched else node.falsy_nodes
        rendered_parts.append(_render_nodes(nodes=branch, context=context))

    return "".join(rendered_parts)


def _resolve_context_value(*, context: dict[str, object], identifier: str) -> object:
    if identifier not in context:
        msg = f"Missing assembly variable '{identifier}'."
        raise ValueError(msg)
    return context[identifier]


def _evaluate_condition(*, condition: _Condition, context: dict[str, object]) -> bool:
    value = _resolve_context_value(context=context, identifier=condition.identifier)
    if condition.expected_literal is None:
        return bool(value)
    return value == condition.expected_literal



from __future__ import annotations

import builtins
import inspect
import typing
from typing import Any, Generic, TypeVar, cast

import pytest

from diwire import BaseScope, Lifetime, LockMode, Scope
from diwire._internal import open_generics
from diwire._internal.providers import ProviderDependency
from diwire.exceptions import (
    DIWireAsyncDependencyInSyncContextError,
    DIWireInvalidGenericTypeArgumentError,
    DIWireScopeMismatchError,
)

T = TypeVar("T")
U = TypeVar("U")


class _Generic(Generic[T]):
    pass


class _Model:
    pass


class _User(_Model):
    pass


M = TypeVar("M", bound=_Model)
N = TypeVar("N", bound=_Model)


def _factory() -> object:
    return object()


def test_canonicalize_and_substitute_handle_aliases_without_arguments() -> None:
    assert open_generics.canonicalize_open_key(typing.Sequence) is None
    assert open_generics.substitute_typevars(typing.Sequence, mapping={}) == typing.Sequence


def test_validate_typevar_arguments_uses_generic_invalid_message_when_check_fails(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setattr(open_generics, "_is_type_argument_valid", lambda **_kwargs: False)

    with pytest.raises(DIWireInvalidGenericTypeArgumentError, match="is invalid"):
        open_generics.validate_typevar_arguments({TypeVar("X"): int})


def test_registry_validation_error_and_no_match_paths() -> None:
    registry = open_generics.OpenGenericRegistry()
    registry.register(
        provides=_Generic[M],
        provider_kind="factory",
        provider=_factory,
        lifetime=Lifetime.TRANSIENT,
        scope=Scope.APP,
        lock_mode=LockMode.NONE,
        is_async=False,
        is_any_dependency_async=False,
        needs_cleanup=False,
        dependencies=[],
    )

    invalid_key = cast("Any", _Generic)[str]
    with pytest.raises(DIWireInvalidGenericTypeArgumentError):
        registry.find_best_match(invalid_key)

    assert registry.find_best_match(list[int]) is None


def test_registry_handles_multiple_invalid_matches_after_first_validation_error() -> None:
    registry = open_generics.OpenGenericRegistry()
    registry.register(
        provides=_Generic[M],
        provider_kind="factory",
        provider=_factory,
        lifetime=Lifetime.TRANSIENT,
        scope=Scope.APP,
        lock_mode=LockMode.NONE,
        is_async=False,
        is_any_dependency_async=False,
        needs_cleanup=False,
        dependencies=[],
    )
    registry.register(
        provides=_Generic[N],
        provider_kind="factory",
        provider=_factory,
        lifetime=Lifetime.TRANSIENT,
        scope=Scope.APP,
        lock_mode=LockMode.NONE,
        is_async=False,
        is_any_dependency_async=False,
        needs_cleanup=False,
        dependencies=[],
    )

    invalid_key = cast("Any", _Generic)[str]
    with pytest.raises(DIWireInvalidGenericTypeArgumentError):
        registry.find_best_match(invalid_key)


def test_append_call_argument_supports_all_parameter_kinds_and_fallback() -> None:
    def _positional_only(value: int, /) -> None:
        _ = value

    def _positional_or_keyword(value: int) -> None:
        _ = value

    def _keyword_only(*, value: int) -> None:
        _ = value

    def _var_positional(*values: int) -> None:
        _ = values

    def _var_keyword(**values: int) -> None:
        _ = values

    positional_arguments: list[Any] = []
    keyword_arguments: dict[str, Any] = {}

    signatures = [
        inspect.signature(_positional_only),
        inspect.signature(_positional_or_keyword),
        inspect.signature(_keyword_only),
        inspect.signature(_var_positional),
        inspect.signature(_var_keyword),
    ]
    dependencies = [
        ProviderDependency(provides=int, parameter=signatures[0].parameters["value"]),
        ProviderDependency(provides=int, parameter=signatures[1].parameters["value"]),
        ProviderDependency(provides=int, parameter=signatures[2].parameters["value"]),
        ProviderDependency(provides=tuple[int, ...], parameter=signatures[3].parameters["values"]),
        ProviderDependency(provides=dict[str, int], parameter=signatures[4].parameters["values"]),
    ]

    open_generics._append_call_argument(
        dependency=dependencies[0],
        value=1,
        positional_arguments=positional_arguments,
        keyword_arguments=keyword_arguments,
    )
    open_generics._append_call_argument(
        dependency=dependencies[1],
        value=2,
        positional_arguments=positional_arguments,
        keyword_arguments=keyword_arguments,
    )
    open_generics._append_call_argument(
        dependency=dependencies[2],
        value=3,
        positional_arguments=positional_arguments,
        keyword_arguments=keyword_arguments,
    )
    open_generics._append_call_argument(
        dependency=dependencies[3],
        value=(4, 5),
        positional_arguments=positional_arguments,
        keyword_arguments=keyword_arguments,
    )
    open_generics._append_call_argument(
        dependency=dependencies[4],
        value={"six": 6},
        positional_arguments=positional_arguments,
        keyword_arguments=keyword_arguments,
    )

    class _UnknownParameter:
        name = "fallback"
        kind = object()

    class _UnknownDependency:
        parameter = _UnknownParameter()

    open_generics._append_call_argument(
        dependency=cast("Any", _UnknownDependency()),
        value=7,
        positional_arguments=positional_arguments,
        keyword_arguments=keyword_arguments,
    )

    assert positional_arguments == [1, 4, 5]
    assert keyword_arguments["value"] == 3
    assert keyword_arguments["six"] == 6
    assert keyword_arguments["fallback"] == 7


def test_cast_iterable_and_cast_mapping_raise_on_invalid_values() -> None:
    with pytest.raises(TypeError, match="Expected iterable value"):
        open_generics.cast_iterable(1)

    with pytest.raises(TypeError, match="Expected mapping value"):
        open_generics.cast_mapping(1)


def test_typevar_collection_matching_and_scoring_helpers_cover_edge_cases() -> None:
    found: list[TypeVar] = []
    open_generics._collect_typevars_into(value=_Generic, found=found)
    assert T in found

    mismatch = open_generics._match_typevars(template=tuple[T, T], concrete=tuple[int, str])
    assert mismatch is None
    assert open_generics._match_typevars(template=int, concrete=int) is not None
    assert open_generics._match_typevars(template=_Generic, concrete=_Generic[int]) is not None
    assert open_generics._match_typevars(template=list[T], concrete=dict[str, int]) is None
    assert open_generics._match_typevars(template=tuple[T, U], concrete=tuple[int]) is None

    assert open_generics._is_closed_generic_dependency(typing.Sequence) is False
    assert open_generics._specificity_score(int) == 2
    assert open_generics._specificity_score(typing.Sequence) == 2
    assert open_generics._normalize_generic_node(_Generic) == _Generic[T]
    assert (
        open_generics._rebuild_alias(origin=object(), args=(int,), fallback="fallback")
        == "fallback"
    )


def test_matches_type_constraint_handles_any_and_issubclass_type_error(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    assert open_generics._matches_type_constraint(argument=int, constraint=Any) is True
    assert open_generics._matches_type_constraint(argument=42, constraint=42) is True

    monkeypatch.setattr(builtins, "issubclass", lambda *_args: (_ for _ in ()).throw(TypeError()))
    assert open_generics._matches_type_constraint(argument=int, constraint=int) is False


def test_resolve_scope_transition_path_handles_all_error_and_success_paths() -> None:
    with pytest.raises(DIWireScopeMismatchError, match="Cannot enter deeper scope"):
        open_generics._resolve_scope_transition_path(
            root_scope=Scope.APP,
            current_scope_level=Scope.STEP.level,
            scope=None,
        )

    assert open_generics._resolve_scope_transition_path(
        root_scope=Scope.APP,
        current_scope_level=Scope.APP.level,
        scope=None,
    ) == [Scope.REQUEST]
    assert open_generics._resolve_scope_transition_path(
        root_scope=Scope.APP,
        current_scope_level=Scope.APP.level,
        scope=Scope.SESSION,
    ) == [Scope.SESSION]
    assert open_generics._resolve_scope_transition_path(
        root_scope=Scope.APP,
        current_scope_level=Scope.APP.level,
        scope=Scope.ACTION,
    ) == [Scope.REQUEST, Scope.ACTION]
    assert (
        open_generics._resolve_scope_transition_path(
            root_scope=Scope.APP,
            current_scope_level=Scope.REQUEST.level,
            scope=Scope.REQUEST,
        )
        == []
    )

    with pytest.raises(DIWireScopeMismatchError, match="Cannot enter scope level"):
        open_generics._resolve_scope_transition_path(
            root_scope=Scope.APP,
            current_scope_level=Scope.REQUEST.level,
            scope=Scope.SESSION,
        )

    with pytest.raises(DIWireScopeMismatchError, match="is not a valid next transition"):
        open_generics._resolve_scope_transition_path(
            root_scope=Scope.APP,
            current_scope_level=Scope.APP.level,
            scope=BaseScope(99),
        )


def test_cleanup_helpers_handle_sync_async_and_error_aggregation() -> None:
    sync_called: list[str] = []
    async_called: list[str] = []

    def _sync_cleanup(
        _exc_type: type[BaseException] | None,
        _exc_value: BaseException | None,
        _traceback: Any,
    ) -> None:
        sync_called.append("sync")

    async def _async_cleanup(
        _exc_type: type[BaseException] | None,
        _exc_value: BaseException | None,
        _traceback: Any,
    ) -> None:
        async_called.append("async")

    open_generics._execute_sync_cleanup_callbacks(
        callbacks=[(0, _sync_cleanup)],
        exc_type=None,
        exc_value=None,
        traceback=None,
    )
    assert sync_called == ["sync"]

    with pytest.raises(DIWireAsyncDependencyInSyncContextError):
        open_generics._execute_sync_cleanup_callbacks(
            callbacks=[(1, _async_cleanup)],
            exc_type=None,
            exc_value=None,
            traceback=None,
        )

    open_generics._execute_sync_cleanup_callbacks(
        callbacks=[(1, _async_cleanup), (0, _sync_cleanup)],
        exc_type=RuntimeError,
        exc_value=RuntimeError(),
        traceback=None,
    )

    async def _run_async_cleanup() -> None:
        await open_generics._execute_async_cleanup_callbacks(
            callbacks=[(0, _sync_cleanup), (1, _async_cleanup)],
            exc_type=None,
            exc_value=None,
            traceback=None,
        )

    import asyncio

    asyncio.run(_run_async_cleanup())
    assert "async" in async_called

    async def _run_async_cleanup_failure() -> None:
        def _failing_cleanup(
            _exc_type: type[BaseException] | None,
            _exc_value: BaseException | None,
            _traceback: Any,
        ) -> None:
            raise ValueError("boom")

        await open_generics._execute_async_cleanup_callbacks(
            callbacks=[(0, _failing_cleanup), (0, _failing_cleanup)],
            exc_type=None,
            exc_value=None,
            traceback=None,
        )

    with pytest.raises(ValueError, match="boom"):
        asyncio.run(_run_async_cleanup_failure())


def test_provider_cast_helpers_and_async_cleanup_error_helper_raise() -> None:
    with pytest.raises(TypeError, match="Expected concrete type provider"):
        open_generics._as_provider_type(1)
    with pytest.raises(TypeError, match="Expected factory provider"):
        open_generics._as_factory_provider(1)
    with pytest.raises(TypeError, match="Expected generator provider"):
        open_generics._as_generator_provider(1)
    with pytest.raises(TypeError, match="Expected context manager provider"):
        open_generics._as_context_manager_provider(1)
    with pytest.raises(DIWireAsyncDependencyInSyncContextError):
        open_generics._raise_async_cleanup_in_sync_context()



"""Register components ergonomically with ``component=...``.

This example shows that registration can use ``component=...`` while runtime
resolution and injection keys remain ``Annotated[..., Component(...)]``.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Annotated

from diwire import Component, Container, Injected, resolver_context


@dataclass(slots=True)
class Cache:
    backend: str


PrimaryCache = Annotated[Cache, Component("primary")]
FallbackCache = Annotated[Cache, Component("fallback")]


def main() -> None:
    container = Container()
    container.add_instance(Cache(backend="redis"), provides=Cache, component="primary")
    container.add_instance(Cache(backend="memory"), provides=Cache, component=Component("fallback"))

    @resolver_context.inject
    def load(
        primary: Injected[PrimaryCache],
        fallback: Injected[FallbackCache],
    ) -> tuple[str, str]:
        return primary.backend, fallback.backend

    primary_backend, fallback_backend = load()
    print(f"primary={primary_backend}")  # => primary=redis
    print(f"fallback={fallback_backend}")  # => fallback=memory
    print(f"resolved={container.resolve(PrimaryCache).backend}")  # => resolved=redis


if __name__ == "__main__":
    main()



"""Focused example: scoped open generics require an opened scope."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Generic, TypeVar

from diwire import Container, Lifetime, Scope
from diwire.exceptions import DIWireScopeMismatchError

T = TypeVar("T")


class IBox(Generic[T]):
    pass


@dataclass(slots=True)
class Box(IBox[T]):
    type_arg: type[T]


def build_box(type_arg: type[T]) -> IBox[T]:
    return Box(type_arg=type_arg)


def main() -> None:
    container = Container()
    container.add_factory(
        build_box,
        provides=IBox,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    try:
        container.resolve(IBox[int])
    except DIWireScopeMismatchError:
        requires_scope = True
    else:
        requires_scope = False

    print(f"scoped_requires_scope={requires_scope}")  # => scoped_requires_scope=True


if __name__ == "__main__":
    main()



"""Focused example: ``add`` for constructor-based creation."""

from __future__ import annotations

from dataclasses import dataclass

from diwire import Container


class Dependency:
    pass


@dataclass(slots=True)
class Service:
    dependency: Dependency


def main() -> None:
    container = Container()
    container.add(Dependency, provides=Dependency)
    container.add(Service, provides=Service)

    resolved = container.resolve(Service)
    print(
        f"concrete_injected_dep={isinstance(resolved.dependency, Dependency)}",
    )  # => concrete_injected_dep=True


if __name__ == "__main__":
    main()



"""Focused example: provider calls follow scoped vs transient lifetime semantics."""

from __future__ import annotations

from diwire import Container, Lifetime, Provider, Scope


class Expensive:
    build_count = 0

    def __init__(self) -> None:
        type(self).build_count += 1


class UsesExpensiveProvider:
    def __init__(self, expensive_provider: Provider[Expensive]) -> None:
        self._expensive_provider = expensive_provider

    def get_expensive(self) -> Expensive:
        return self._expensive_provider()


def _run_scenario(*, lifetime: Lifetime) -> tuple[int, bool]:
    Expensive.build_count = 0
    container = Container()
    container.add(
        Expensive,
        provides=Expensive,
        scope=Scope.REQUEST,
        lifetime=lifetime,
    )
    container.add(
        UsesExpensiveProvider,
        provides=UsesExpensiveProvider,
        scope=Scope.REQUEST,
        lifetime=Lifetime.TRANSIENT,
    )

    with container.enter_scope() as request_scope:
        consumer = request_scope.resolve(UsesExpensiveProvider)
        first = consumer.get_expensive()
        second = consumer.get_expensive()

    return Expensive.build_count, first is second


def main() -> None:
    scoped_calls, scoped_same = _run_scenario(lifetime=Lifetime.SCOPED)
    transient_calls, transient_same = _run_scenario(lifetime=Lifetime.TRANSIENT)

    print(f"scoped_after_calls={scoped_calls}")  # => scoped_after_calls=1
    print(f"scoped_same_identity={scoped_same}")  # => scoped_same_identity=True
    print(f"transient_after_calls={transient_calls}")  # => transient_after_calls=2
    print(f"transient_same_identity={transient_same}")  # => transient_same_identity=False


if __name__ == "__main__":
    main()



"""Injected callables can consume FromContext values via diwire_context."""

from __future__ import annotations

from diwire import Container, FromContext, Scope, resolver_context


def main() -> None:
    Container()

    @resolver_context.inject(scope=Scope.REQUEST)
    def handler(value: FromContext[int]) -> int:
        return value

    from_context = handler(diwire_context={int: 7})
    overridden = handler(value=8)

    print(f"from_context={from_context}")  # => from_context=7
    print(f"overridden={overridden}")  # => overridden=8


if __name__ == "__main__":
    main()



from __future__ import annotations

import functools
import inspect
from collections.abc import Callable, Iterator
from contextlib import suppress
from typing import Any, cast

import pytest

from diwire._internal.container import Container
from diwire._internal.injection import (
    INJECT_RESOLVER_KWARG,
    InjectedCallableInspector,
    InjectedParameter,
)
from diwire._internal.resolver_context import resolver_context

_DIWIRE_CONTAINER_ATTR = "_diwire_container"
_DIWIRE_INJECTED_PARAMETERS_ATTR = "__diwire_pytest_injected_parameters__"
_DIWIRE_ORIGINAL_SIGNATURE_ATTR = "__diwire_pytest_original_signature__"
_INJECTED_CALLABLE_INSPECTOR = InjectedCallableInspector()


@pytest.fixture()
def diwire_container() -> Container:
    """Create a per-test DI container used by the plugin.

    Tests that use ``Injected[...]`` parameters resolve them from this container.
    The fixture is function-scoped, so registrations are isolated between tests
    unless users override fixture scope explicitly.

    Returns:
        A new ``Container`` instance.

    """
    return Container()


@pytest.fixture(autouse=True)
def _diwire_state(
    request: pytest.FixtureRequest,
    diwire_container: Container,
) -> None:
    """Store plugin state on the test node for hook access."""
    node = cast("Any", request.node)
    setattr(node, _DIWIRE_CONTAINER_ATTR, diwire_container)


def pytest_pycollect_makeitem(
    collector: Any,
    name: str,
    obj: object,
) -> Any | None:
    """Hide ``Injected[...]`` parameters from pytest fixture name matching.

    Pytest treats test function parameters as fixture names. This hook rewrites
    signatures for callables that use DIWire injection metadata so injected
    parameters are not interpreted as missing fixtures.

    Args:
        collector: Pytest collector instance.
        name: Collected object name.
        obj: Candidate object.

    Returns:
        ``None`` to continue default collection flow.

    """
    if not callable(obj):
        return None
    if not collector.istestfunction(obj, name):
        return None

    callable_obj = cast("Callable[..., Any]", obj)
    inspection = _INJECTED_CALLABLE_INSPECTOR.inspect_callable(callable_obj)
    if not inspection.injected_parameters:
        return None

    obj_as_any = cast("Any", obj)
    obj_as_any.__dict__[_DIWIRE_INJECTED_PARAMETERS_ATTR] = inspection.injected_parameters
    obj_as_any.__dict__[_DIWIRE_ORIGINAL_SIGNATURE_ATTR] = inspection.signature
    obj_as_any.__signature__ = inspection.public_signature
    return None


@pytest.hookimpl(hookwrapper=True, tryfirst=True)
def pytest_pyfunc_call(pyfuncitem: pytest.Function) -> Iterator[None]:
    """Wrap test function execution to resolve ``Injected[...]`` parameters.

    The wrapper swaps the callable with ``resolver_context.inject(original_callable)``
    for the duration of the test call and executes with an explicit resolver
    kwarg. If no container state is attached to the node, this hook is a no-op.

    Limitations:
        Injection requires a callable test object and plugin-managed state on the
        test item. Non-function collection paths are ignored.

    Args:
        pyfuncitem: Collected pytest function item.

    Yields:
        Control back to pytest around test execution.

    """
    original_callable = cast("Callable[..., Any]", pyfuncitem.obj)
    original_callable_as_any = cast("Any", original_callable)
    injected_parameters = cast(
        "tuple[InjectedParameter, ...] | None",
        getattr(original_callable_as_any, _DIWIRE_INJECTED_PARAMETERS_ATTR, None),
    )
    if injected_parameters is None:
        injected_parameters = _INJECTED_CALLABLE_INSPECTOR.inspect_callable(
            original_callable,
        ).injected_parameters
    if not injected_parameters:
        yield
        return

    item = cast("Any", pyfuncitem)
    container = cast("Container | None", getattr(item, _DIWIRE_CONTAINER_ATTR, None))
    if container is None:
        yield
        return

    had_signature_override = hasattr(original_callable_as_any, "__signature__")
    signature_override = cast("Any", getattr(original_callable_as_any, "__signature__", None))
    original_signature = cast(
        "inspect.Signature | None",
        getattr(original_callable_as_any, _DIWIRE_ORIGINAL_SIGNATURE_ATTR, None),
    )
    if original_signature is not None:
        original_callable_as_any.__signature__ = original_signature

    try:
        injected_callable = resolver_context.inject(original_callable)
        resolver = container.compile()
        if inspect.iscoroutinefunction(injected_callable):

            @functools.wraps(injected_callable)
            async def _invoke_with_resolver(*args: Any, **kwargs: Any) -> Any:
                kwargs[INJECT_RESOLVER_KWARG] = resolver
                return await cast("Any", injected_callable)(*args, **kwargs)

        else:

            @functools.wraps(injected_callable)
            def _invoke_with_resolver(*args: Any, **kwargs: Any) -> Any:
                kwargs[INJECT_RESOLVER_KWARG] = resolver
                return injected_callable(*args, **kwargs)

        pyfuncitem.obj = _invoke_with_resolver
    finally:
        if had_signature_override:
            original_callable_as_any.__signature__ = signature_override
        else:
            with suppress(AttributeError):
                del original_callable_as_any.__signature__

    try:
        yield
    finally:
        pyfuncitem.obj = original_callable



from __future__ import annotations

import inspect
import re
from collections.abc import AsyncGenerator, Generator
from contextlib import asynccontextmanager
from pathlib import Path

from diwire import BaseScope, Container, FromContext, Injected, Lifetime, Scope, resolver_context
from diwire._internal.providers import ProviderSpec
from diwire._internal.resolvers.assembly.renderer import ResolversAssemblyRenderer

_EXPECTED_DIR = Path(__file__).with_name("assembly_expected")


class _SnapshotSession:
    pass


class _SnapshotRequest:
    def __init__(self, session: _SnapshotSession) -> None:
        self.session = session


async def _provide_int_for_snapshot() -> int:
    return 42


class _SnapshotGeneratorResource:
    pass


def _provide_sync_generator_resource() -> Generator[_SnapshotGeneratorResource, None, None]:
    yield _SnapshotGeneratorResource()


class _SnapshotAsyncContextResource:
    pass


@asynccontextmanager
async def _provide_async_context_resource() -> AsyncGenerator[_SnapshotAsyncContextResource, None]:
    yield _SnapshotAsyncContextResource()


class _SnapshotMixedShapeService:
    def __init__(
        self,
        positional: int,
        values: tuple[int, ...],
        options: dict[str, int],
    ) -> None:
        self.positional = positional
        self.values = values
        self.options = options


def _build_snapshot_mixed_shape_service(
    positional: int,
    /,
    *values: int,
    **options: int,
) -> _SnapshotMixedShapeService:
    return _SnapshotMixedShapeService(positional=positional, values=tuple(values), options=options)


class _SnapshotRequestRootAppService:
    pass


class _SnapshotRequestRootSessionService:
    pass


class _SnapshotRequestRootRequestService:
    pass


class _SnapshotActionRootRequestService:
    pass


class _SnapshotActionRootActionService:
    pass


class _SnapshotActionRootStepService:
    pass


class _SnapshotAsyncCleanupSignatureService:
    def __init__(
        self,
        dependency: _SnapshotAsyncContextResource,
        values: tuple[int, ...],
        options: dict[str, int],
    ) -> None:
        self.dependency = dependency
        self.values = values
        self.options = options


def _build_snapshot_async_cleanup_signature_service(
    dependency: _SnapshotAsyncContextResource,
    /,
    *values: int,
    **options: int,
) -> _SnapshotAsyncCleanupSignatureService:
    return _SnapshotAsyncCleanupSignatureService(
        dependency=dependency,
        values=tuple(values),
        options=options,
    )


class _SnapshotInjectWrapperDependency:
    pass


class _SnapshotInjectWrapperService:
    def __init__(self, dependency: _SnapshotInjectWrapperDependency) -> None:
        self.dependency = dependency


def _build_snapshot_inject_wrapper_service(
    dependency: Injected[_SnapshotInjectWrapperDependency],
) -> _SnapshotInjectWrapperService:
    return _SnapshotInjectWrapperService(dependency=dependency)


class _SnapshotAsyncInjectWrapperDependency:
    pass


class _SnapshotAsyncInjectWrapperService:
    def __init__(self, dependency: _SnapshotAsyncInjectWrapperDependency) -> None:
        self.dependency = dependency


async def _build_snapshot_async_inject_wrapper_service(
    dependency: Injected[_SnapshotAsyncInjectWrapperDependency],
) -> _SnapshotAsyncInjectWrapperService:
    return _SnapshotAsyncInjectWrapperService(dependency=dependency)


class _SnapshotInlineRootInjectDependency:
    pass


class _SnapshotInlineRootInjectService:
    def __init__(self, dependency: _SnapshotInlineRootInjectDependency) -> None:
        self.dependency = dependency


class _SnapshotInlineRootRequestService:
    def __init__(self, dependency: _SnapshotInlineRootInjectService) -> None:
        self.dependency = dependency


def _build_snapshot_inline_root_inject_service(
    dependency: Injected[_SnapshotInlineRootInjectDependency],
) -> _SnapshotInlineRootInjectService:
    return _SnapshotInlineRootInjectService(dependency=dependency)


def _build_snapshot_inline_root_request_service(
    dependency: _SnapshotInlineRootInjectService,
) -> _SnapshotInlineRootRequestService:
    return _SnapshotInlineRootRequestService(dependency=dependency)


class _SnapshotInjectWrapperVarKwService:
    def __init__(self, options: dict[str, int]) -> None:
        self.options = options


def _build_snapshot_inject_wrapper_varkw_service(
    **options: int,
) -> _SnapshotInjectWrapperVarKwService:
    return _SnapshotInjectWrapperVarKwService(options=options)


class _SnapshotFromContextService:
    def __init__(self, value: int) -> None:
        self.value = value


def _build_snapshot_from_context_service(
    value: FromContext[int],
) -> _SnapshotFromContextService:
    return _SnapshotFromContextService(value=value)


def test_assembly_matches_expected_for_empty_app_root_graph() -> None:
    ProviderSpec.SLOT_COUNTER = 0
    container = Container()
    generated = _render(container=container, root_scope=Scope.APP)
    expected = _read_expected("app_root_empty.txt")
    assert _normalize_dynamic_metadata(generated) == _normalize_dynamic_metadata(expected)


def test_assembly_matches_expected_for_scoped_graph() -> None:
    ProviderSpec.SLOT_COUNTER = 0
    container = Container()
    container.add(
        _SnapshotSession,
        provides=_SnapshotSession,
        scope=Scope.SESSION,
        lifetime=Lifetime.SCOPED,
    )
    container.add(
        _SnapshotRequest,
        provides=_SnapshotRequest,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    generated = _render(container=container, root_scope=Scope.APP)
    expected = _read_expected("app_root_scoped.txt")
    assert _normalize_dynamic_metadata(generated) == _normalize_dynamic_metadata(expected)


def test_assembly_matches_expected_for_async_graph() -> None:
    ProviderSpec.SLOT_COUNTER = 0
    container = Container()
    container.add_factory(
        _provide_int_for_snapshot,
        provides=int,
        lifetime=Lifetime.SCOPED,
    )
    generated = _render(container=container, root_scope=Scope.APP)
    expected = _read_expected("app_root_async.txt")
    assert _normalize_dynamic_metadata(generated) == _normalize_dynamic_metadata(expected)


def test_assembly_matches_expected_for_sync_generator_graph() -> None:
    ProviderSpec.SLOT_COUNTER = 0
    container = Container()
    container.add_generator(
        _provide_sync_generator_resource,
        provides=_SnapshotGeneratorResource,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    generated = _render(container=container, root_scope=Scope.APP)
    expected = _read_expected("app_root_sync_generator.txt")
    assert _normalize_dynamic_metadata(generated) == _normalize_dynamic_metadata(expected)


def test_assembly_matches_expected_for_async_context_manager_graph() -> None:
    ProviderSpec.SLOT_COUNTER = 0
    container = Container()
    container.add_context_manager(
        _provide_async_context_resource,
        provides=_SnapshotAsyncContextResource,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    generated = _render(container=container, root_scope=Scope.APP)
    expected = _read_expected("app_root_async_context_manager.txt")
    assert _normalize_dynamic_metadata(generated) == _normalize_dynamic_metadata(expected)


def test_assembly_matches_expected_for_mixed_dependency_shape_graph() -> None:
    signature = inspect.signature(_build_snapshot_mixed_shape_service)
    positional_type = int
    values_type = tuple[int, ...]
    options_type = dict[str, int]

    ProviderSpec.SLOT_COUNTER = 0
    container = Container()
    container.add_instance(1, provides=positional_type)
    container.add_instance((2, 3), provides=values_type)
    container.add_instance({"first": 1, "second": 2}, provides=options_type)
    container.add_factory(
        _build_snapshot_mixed_shape_service,
        provides=_SnapshotMixedShapeService,
        dependencies={
            positional_type: signature.parameters["positional"],
            values_type: signature.parameters["values"],
            options_type: signature.parameters["options"],
        },
    )
    generated = _render(container=container, root_scope=Scope.APP)
    expected = _read_expected("app_root_mixed_dependency_shapes.txt")
    assert _normalize_dynamic_metadata(generated) == _normalize_dynamic_metadata(expected)


def test_assembly_matches_expected_for_request_root_filtered_graph() -> None:
    ProviderSpec.SLOT_COUNTER = 0
    container = Container()
    container.add(
        _SnapshotRequestRootAppService,
        provides=_SnapshotRequestRootAppService,
        scope=Scope.APP,
        lifetime=Lifetime.SCOPED,
    )
    container.add(
        _SnapshotRequestRootSessionService,
        provides=_SnapshotRequestRootSessionService,
        scope=Scope.SESSION,
        lifetime=Lifetime.SCOPED,
    )
    container.add(
        _SnapshotRequestRootRequestService,
        provides=_SnapshotRequestRootRequestService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    generated = _render(container=container, root_scope=Scope.REQUEST)
    expected = _read_expected("request_root_filtered.txt")
    assert _normalize_dynamic_metadata(generated) == _normalize_dynamic_metadata(expected)


def test_assembly_matches_expected_for_action_root_filtered_graph() -> None:
    ProviderSpec.SLOT_COUNTER = 0
    container = Container()
    container.add(
        _SnapshotActionRootRequestService,
        provides=_SnapshotActionRootRequestService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    container.add(
        _SnapshotActionRootActionService,
        provides=_SnapshotActionRootActionService,
        scope=Scope.ACTION,
        lifetime=Lifetime.SCOPED,
    )
    container.add(
        _SnapshotActionRootStepService,
        provides=_SnapshotActionRootStepService,
        scope=Scope.STEP,
        lifetime=Lifetime.SCOPED,
    )
    generated = _render(container=container, root_scope=Scope.ACTION)
    expected = _read_expected("action_root_filtered.txt")
    assert _normalize_dynamic_metadata(generated) == _normalize_dynamic_metadata(expected)


def test_assembly_matches_expected_for_async_cleanup_mixed_signature_graph() -> None:
    signature = inspect.signature(_build_snapshot_async_cleanup_signature_service)
    values_type = tuple[int, ...]
    options_type = dict[str, int]

    ProviderSpec.SLOT_COUNTER = 0
    container = Container()
    container.add_context_manager(
        _provide_async_context_resource,
        provides=_SnapshotAsyncContextResource,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )
    container.add_instance((2, 3), provides=values_type)
    container.add_instance({"first": 1, "second": 2}, provides=options_type)
    container.add_factory(
        _build_snapshot_async_cleanup_signature_service,
        provides=_SnapshotAsyncCleanupSignatureService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
        dependencies={
            _SnapshotAsyncContextResource: signature.parameters["dependency"],
            values_type: signature.parameters["values"],
            options_type: signature.parameters["options"],
        },
    )
    generated = _render(container=container, root_scope=Scope.APP)
    expected = _read_expected("app_root_async_cleanup_signature_mixed.txt")
    assert _normalize_dynamic_metadata(generated) == _normalize_dynamic_metadata(expected)


def test_assembly_matches_expected_for_inject_wrapper_provider_graph() -> None:
    ProviderSpec.SLOT_COUNTER = 0
    container = Container()
    container.add_instance(
        _SnapshotInjectWrapperDependency(),
        provides=_SnapshotInjectWrapperDependency,
    )

    build_service = resolver_context.inject(_build_snapshot_inject_wrapper_service)

    container.add_factory(
        build_service,
        provides=_SnapshotInjectWrapperService,
        lifetime=Lifetime.TRANSIENT,
    )
    generated = _render(container=container, root_scope=Scope.APP)
    expected = _read_expected("app_root_inject_wrapper_provider.txt")
    assert _normalize_dynamic_metadata(generated) == _normalize_dynamic_metadata(expected)


def test_assembly_matches_expected_for_async_inject_wrapper_provider_graph() -> None:
    ProviderSpec.SLOT_COUNTER = 0
    container = Container()
    container.add_instance(
        _SnapshotAsyncInjectWrapperDependency(),
        provides=_SnapshotAsyncInjectWrapperDependency,
    )

    build_service = resolver_context.inject(_build_snapshot_async_inject_wrapper_service)

    container.add_factory(
        build_service,
        provides=_SnapshotAsyncInjectWrapperService,
        lifetime=Lifetime.TRANSIENT,
    )
    generated = _render(container=container, root_scope=Scope.APP)
    expected = _read_expected("app_root_inject_wrapper_provider_async.txt")
    assert _normalize_dynamic_metadata(generated) == _normalize_dynamic_metadata(expected)


def test_assembly_matches_expected_for_nested_inline_root_inject_wrapper_graph() -> None:
    ProviderSpec.SLOT_COUNTER = 0
    container = Container()
    container.add_instance(
        _SnapshotInlineRootInjectDependency(),
        provides=_SnapshotInlineRootInjectDependency,
    )

    build_inject_service = resolver_context.inject(_build_snapshot_inline_root_inject_service)
    container.add_factory(
        build_inject_service,
        provides=_SnapshotInlineRootInjectService,
        lifetime=Lifetime.TRANSIENT,
    )
    container.add_factory(
        _build_snapshot_inline_root_request_service,
        provides=_SnapshotInlineRootRequestService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.TRANSIENT,
    )

    generated = _render(container=container, root_scope=Scope.APP)
    expected = _read_expected("app_root_inject_wrapper_nested_inline_root.txt")
    assert _normalize_dynamic_metadata(generated) == _normalize_dynamic_metadata(expected)


def test_assembly_matches_expected_for_inject_wrapper_varkw_argument_order_graph() -> None:
    signature = inspect.signature(_build_snapshot_inject_wrapper_varkw_service)
    options_type = dict[str, int]

    ProviderSpec.SLOT_COUNTER = 0
    container = Container()
    container.add_instance({"first": 1, "second": 2}, provides=options_type)

    build_service = resolver_context.inject(_build_snapshot_inject_wrapper_varkw_service)

    container.add_factory(
        build_service,
        provides=_SnapshotInjectWrapperVarKwService,
        lifetime=Lifetime.TRANSIENT,
        dependencies={
            options_type: signature.parameters["options"],
        },
    )

    generated = _render(container=container, root_scope=Scope.APP)
    expected = _read_expected("app_root_inject_wrapper_varkw_order.txt")
    assert _normalize_dynamic_metadata(generated) == _normalize_dynamic_metadata(expected)


def test_assembly_matches_expected_for_from_context_dependency_graph() -> None:
    ProviderSpec.SLOT_COUNTER = 0
    container = Container()
    container.add_factory(
        _build_snapshot_from_context_service,
        provides=_SnapshotFromContextService,
        scope=Scope.REQUEST,
        lifetime=Lifetime.TRANSIENT,
    )

    generated = _render(container=container, root_scope=Scope.APP)
    expected = _read_expected("app_root_from_context_provider.txt")
    assert _normalize_dynamic_metadata(generated) == _normalize_dynamic_metadata(expected)


def test_normalize_dynamic_metadata_normalizes_annotated_symbol_variants() -> None:
    short_symbol = "Dependency wiring: value (positional_or_keyword) -> context [Annotated]\n"
    qualified_symbol = (
        "Dependency wiring: value (positional_or_keyword) -> context [typing.Annotated]\n"
    )

    normalized_short_symbol = _normalize_dynamic_metadata(short_symbol)
    normalized_qualified_symbol = _normalize_dynamic_metadata(qualified_symbol)

    assert normalized_short_symbol == normalized_qualified_symbol
    assert (
        normalized_short_symbol
        == "Dependency wiring: value (positional_or_keyword) -> context [typing.Annotated]\n"
    )


def _render(*, container: Container, root_scope: BaseScope) -> str:
    renderer = ResolversAssemblyRenderer()
    return renderer.get_providers_code(
        root_scope=root_scope,
        registrations=container._providers_registrations,
    )


def _read_expected(file_name: str) -> str:
    return (_EXPECTED_DIR / file_name).read_text()


def _normalize_dynamic_metadata(code: str) -> str:
    normalized = re.sub(
        r"diwire version used for generation: .*",
        "diwire version used for generation: <dynamic>",
        code,
    )
    return re.sub(r"\[(?:typing\.)?Annotated\]", "[typing.Annotated]", normalized)



"""Focused example: ``Injected[T]`` and public signature filtering."""

from __future__ import annotations

import inspect
from dataclasses import dataclass

from diwire import Container, Injected, resolver_context


@dataclass(slots=True)
class User:
    email: str


def main() -> None:
    container = Container()
    container.add_instance(User(email="user@example.com"))

    @resolver_context.inject
    def handler(user_email: str, user: Injected[User], user_name: str) -> str:
        return f"{user_email}|{user_name}|{user.email}"

    signature = "('" + "','".join(inspect.signature(handler).parameters) + "')"
    print(f"signature={signature}")  # => signature=('user_email','user_name')


if __name__ == "__main__":
    main()



"""Passing diwire_context without opening a scope raises a clear error."""

from __future__ import annotations

from diwire import Container, FromContext, resolver_context
from diwire.exceptions import DIWireInvalidRegistrationError


def main() -> None:
    Container()

    @resolver_context.inject(auto_open_scope=False)
    def handler(value: FromContext[int]) -> int:
        return value

    try:
        handler(diwire_context={int: 7})
    except DIWireInvalidRegistrationError as error:
        error_name = type(error).__name__

    print(
        f"context_without_scope_error={error_name}"
    )  # => context_without_scope_error=DIWireInvalidRegistrationError


if __name__ == "__main__":
    main()



from typing import Any, cast

from diwire._internal.providers import ProvidersRegistrations
from diwire._internal.resolvers.assembly.planner import validate_resolver_assembly_managed_scopes
from diwire._internal.resolvers.assembly.renderer import ResolversAssemblyRenderer
from diwire._internal.resolvers.protocol import BuildRootResolverFunctionProtocol, ResolverProtocol
from diwire._internal.scope import BaseScope


class ResolversManager:
    """Manager for dependency resolvers."""

    def __init__(self) -> None:
        self._assembly_renderer = ResolversAssemblyRenderer()

    def build_root_resolver(
        self,
        root_scope: BaseScope,
        registrations: ProvidersRegistrations,
    ) -> ResolverProtocol:
        """Get the root resolver for the given registrations.

        Generates the resolver code dynamically based on the provided registrations and root scope.

        Args:
            root_scope: Root scope used to initialize the resolver.
            registrations: Provider registrations used to build resolver instances or generated code.

        """
        validate_resolver_assembly_managed_scopes(root_scope=root_scope)
        code = self._assembly_renderer.get_providers_code(
            root_scope=root_scope,
            registrations=registrations,
        )
        managed_scopes = validate_resolver_assembly_managed_scopes(root_scope=root_scope)

        namespace: dict[str, Any] = {}
        exec(code, namespace)  # noqa: S102

        for scope in managed_scopes:
            scope_binding_name = f"_scope_obj_{scope.level}"
            if scope_binding_name in namespace:
                namespace[scope_binding_name] = scope

        build_root_resolver = cast(
            "BuildRootResolverFunctionProtocol",
            namespace["build_root_resolver"],
        )

        return build_root_resolver(registrations)



from __future__ import annotations

import pytest

from diwire import Container, FromContext, Lifetime, Scope
from diwire.exceptions import DIWireDependencyNotRegisteredError, DIWireScopeMismatchError


class _RequestOnly:
    pass


class _ActionFromContext:
    def __init__(self, value: FromContext[int]) -> None:
        self.value = value


class _AsyncRequestOnly:
    pass


def test_container_resolve_uses_context_bound_scope_by_default() -> None:
    container = Container()
    container.add(
        _RequestOnly,
        provides=_RequestOnly,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope(Scope.REQUEST):
        assert isinstance(container.resolve(_RequestOnly), _RequestOnly)


@pytest.mark.asyncio
async def test_container_aresolve_uses_context_bound_scope_by_default() -> None:
    container = Container()

    async def build() -> _AsyncRequestOnly:
        return _AsyncRequestOnly()

    container.add_factory(
        build,
        provides=_AsyncRequestOnly,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope(Scope.REQUEST):
        assert isinstance(await container.aresolve(_AsyncRequestOnly), _AsyncRequestOnly)


def test_container_resolve_does_not_use_context_bound_scope_when_disabled() -> None:
    container = Container(use_resolver_context=False)
    container.add(
        _RequestOnly,
        provides=_RequestOnly,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope(Scope.REQUEST):
        with pytest.raises(DIWireScopeMismatchError, match="requires opened scope level"):
            _ = container.resolve(_RequestOnly)


def test_container_enter_scope_uses_context_bound_scope_for_nesting() -> None:
    container = Container()
    container.add(
        _ActionFromContext,
        provides=_ActionFromContext,
        scope=Scope.ACTION,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope(Scope.REQUEST, context={int: 7}):
        with container.enter_scope(Scope.ACTION):
            assert container.resolve(_ActionFromContext).value == 7


def test_container_enter_scope_without_nesting_does_not_inherit_context() -> None:
    container = Container(use_resolver_context=False)
    container.add(
        _ActionFromContext,
        provides=_ActionFromContext,
        scope=Scope.ACTION,
        lifetime=Lifetime.SCOPED,
    )

    with container.enter_scope(Scope.REQUEST, context={int: 7}):
        with container.enter_scope(Scope.ACTION) as action_resolver:
            with pytest.raises(DIWireDependencyNotRegisteredError, match="Context value"):
                _ = action_resolver.resolve(_ActionFromContext)



from __future__ import annotations

import importlib
import pathlib
import uuid
import warnings
from abc import ABC, abstractmethod
from types import ModuleType
from typing import Annotated, Any, Generic, NamedTuple, TypeVar, cast

import msgspec
import pytest

from diwire import (
    AsyncProvider,
    Container,
    DependencyRegistrationPolicy,
    Lifetime,
    MissingPolicy,
    Provider,
    Scope,
)
from diwire._internal.autoregistration import ConcreteTypeAutoregistrationPolicy
from diwire.exceptions import DIWireDependencyNotRegisteredError

_PYDANTIC_V1_WARNING_PATTERN = (
    r"Core Pydantic V1 functionality isn't compatible with Python 3\.14 or greater\."
)


def _load_pydantic_v1() -> ModuleType | None:
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=_PYDANTIC_V1_WARNING_PATTERN,
            category=UserWarning,
        )
        try:
            return importlib.import_module("pydantic.v1")
        except ImportError:
            return None


pydantic_v1 = _load_pydantic_v1()


def _load_pydantic_settings_base() -> type[Any] | None:
    try:
        module = importlib.import_module("pydantic_settings")
    except ImportError:
        return None
    base_settings = getattr(module, "BaseSettings", None)
    if isinstance(base_settings, type):
        return cast("type[Any]", base_settings)
    return None


pydantic_settings_base = _load_pydantic_settings_base()

_PydanticSettingsBase: type[Any]
if pydantic_settings_base is None:

    class _MissingPydanticSettingsBase:
        pass

    _PydanticSettingsBase = _MissingPydanticSettingsBase
else:
    _PydanticSettingsBase = pydantic_settings_base


class DirectDependency:
    pass


class RootWithDirectDependency:
    def __init__(self, dependency: DirectDependency) -> None:
        self.dependency = dependency


class NestedDependency:
    pass


class IntermediateDependency:
    def __init__(self, dependency: NestedDependency) -> None:
        self.dependency = dependency


class RootWithNestedDependencies:
    def __init__(self, dependency: IntermediateDependency) -> None:
        self.dependency = dependency


class ScopedDependency:
    pass


class RootWithScopedDependency:
    def __init__(self, dependency: ScopedDependency) -> None:
        self.dependency = dependency


class AbstractDependency(ABC):
    @abstractmethod
    def run(self) -> None:
        """Run the dependency."""


class ValidDependency:
    pass


class RootWithMixedDependencies:
    def __init__(
        self,
        abstract_dependency: AbstractDependency,
        valid_dependency: ValidDependency,
    ) -> None:
        self.abstract_dependency = abstract_dependency
        self.valid_dependency = valid_dependency


class DecoratorDependency:
    pass


class DecoratorService:
    def __init__(self, dependency: DecoratorDependency) -> None:
        self.dependency = dependency


class ResolveDependency:
    pass


class ResolveRoot:
    def __init__(self, dependency: ResolveDependency) -> None:
        self.dependency = dependency


class ExistingDependency:
    pass


class RootWithExistingDependency:
    def __init__(self, dependency: ExistingDependency) -> None:
        self.dependency = dependency


class NamedTupleFrameworkDependency:
    pass


class MsgspecFrameworkDependency:
    pass


class PydanticFrameworkDependency:
    pass


class PydanticSettingsDependency(_PydanticSettingsBase):
    value: str = "settings"


class RootWithPydanticSettingsDependency:
    def __init__(self, dependency: PydanticSettingsDependency) -> None:
        self.dependency = dependency


class RootWithBuiltinDep:
    def __init__(self, token: str) -> None:
        self.token = token


class RootWithPathDep:
    def __init__(self, path: pathlib.Path) -> None:
        self.path = path


class RootWithUuidDep:
    def __init__(self, request_id: uuid.UUID) -> None:
        self.request_id = request_id


class _AutoProviderDependency:
    pass


_OpenT = TypeVar("_OpenT")


class _OpenAutoregDependency(Generic[_OpenT]):
    pass


def _build_open_autoreg_dependency(type_arg: type[_OpenT]) -> _OpenAutoregDependency[_OpenT]:
    return _OpenAutoregDependency()


class _TestMetaclass(type):
    pass


def test_container_default_autoregisters_dependencies_during_registration() -> None:
    container = Container()

    container.add(RootWithDirectDependency)

    assert container._providers_registrations.find_by_type(DirectDependency) is not None


def test_registration_override_can_disable_dependency_autoregistration() -> None:
    container = Container(missing_policy=MissingPolicy.REGISTER_ROOT)

    container.add(
        RootWithDirectDependency,
        dependency_registration_policy=DependencyRegistrationPolicy.IGNORE,
    )

    assert container._providers_registrations.find_by_type(DirectDependency) is None


def test_registration_override_can_enable_dependency_autoregistration() -> None:
    container = Container(dependency_registration_policy=DependencyRegistrationPolicy.IGNORE)

    container.add(
        RootWithDirectDependency,
        dependency_registration_policy=DependencyRegistrationPolicy.REGISTER_RECURSIVE,
    )

    assert container._providers_registrations.find_by_type(DirectDependency) is not None


def test_dependency_autoregistration_is_recursive() -> None:
    container = Container(
        dependency_registration_policy=DependencyRegistrationPolicy.REGISTER_RECURSIVE
    )

    container.add(RootWithNestedDependencies)

    assert container._providers_registrations.find_by_type(IntermediateDependency) is not None
    assert container._providers_registrations.find_by_type(NestedDependency) is not None


def test_dependency_autoregistration_skips_already_registered_dependency() -> None:
    container = Container(
        dependency_registration_policy=DependencyRegistrationPolicy.REGISTER_RECURSIVE
    )
    container.add(ExistingDependency)
    existing_spec = container._providers_registrations.get_by_type(ExistingDependency)

    container.add(RootWithExistingDependency)

    assert container._providers_registrations.get_by_type(ExistingDependency) is existing_spec


def test_dependency_registration_policy_recursive_autoregisters_dependencies() -> None:
    container = Container(
        dependency_registration_policy=DependencyRegistrationPolicy.REGISTER_RECURSIVE,
    )

    container.add(RootWithDirectDependency)

    assert container._providers_registrations.find_by_type(DirectDependency) is not None


def test_dependency_autoregistration_uses_parent_scope_and_lifetime() -> None:
    container = Container(
        dependency_registration_policy=DependencyRegistrationPolicy.REGISTER_RECURSIVE
    )

    container.add(
        RootWithScopedDependency,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
        dependency_registration_policy=DependencyRegistrationPolicy.REGISTER_RECURSIVE,
    )

    dependency_spec = container._providers_registrations.get_by_type(ScopedDependency)
    assert dependency_spec.scope is Scope.REQUEST
    assert dependency_spec.lifetime is Lifetime.SCOPED


def test_dependency_autoregistration_ignores_registration_failures_and_continues() -> None:
    container = Container(
        dependency_registration_policy=DependencyRegistrationPolicy.REGISTER_RECURSIVE
    )

    container.add(RootWithMixedDependencies)

    assert container._providers_registrations.find_by_type(AbstractDependency) is None
    assert container._providers_registrations.find_by_type(ValidDependency) is not None


def test_resolve_autoregistration_skips_builtin_dependency_types() -> None:
    container = Container(missing_policy=MissingPolicy.REGISTER_ROOT)

    with pytest.raises(DIWireDependencyNotRegisteredError) as exc_info:
        container.resolve(RootWithBuiltinDep)

    message = str(exc_info.value)
    assert repr(str) in message
    assert repr(RootWithBuiltinDep) in message
    assert container._providers_registrations.find_by_type(str) is None


def test_resolve_autoregistration_skips_pathlib_path_dependency() -> None:
    container = Container(missing_policy=MissingPolicy.REGISTER_ROOT)

    with pytest.raises(DIWireDependencyNotRegisteredError):
        container.resolve(RootWithPathDep)

    assert container._providers_registrations.find_by_type(pathlib.Path) is None
    expected_path = pathlib.Path("x")
    container.add_instance(
        expected_path,
        provides=pathlib.Path,
    )

    resolved = container.resolve(RootWithPathDep)
    assert resolved.path is expected_path


def test_resolve_autoregistration_skips_uuid_dependency() -> None:
    container = Container(missing_policy=MissingPolicy.REGISTER_ROOT)

    with pytest.raises(DIWireDependencyNotRegisteredError):
        container.resolve(RootWithUuidDep)

    assert container._providers_registrations.find_by_type(uuid.UUID) is None
    expected_uuid = uuid.UUID(int=0)
    container.add_instance(expected_uuid)

    resolved = container.resolve(RootWithUuidDep)
    assert resolved.request_id is expected_uuid


def test_autoregistration_policy_skips_metaclasses() -> None:
    policy = ConcreteTypeAutoregistrationPolicy()

    assert policy.is_eligible_concrete(_TestMetaclass) is False


def test_autoregistration_policy_skips_non_runtime_class_candidates() -> None:
    policy = ConcreteTypeAutoregistrationPolicy()

    assert policy.is_eligible_concrete("not-a-class") is False


def test_factory_registration_can_set_dependency_registration_policy() -> None:
    container = Container(
        dependency_registration_policy=DependencyRegistrationPolicy.REGISTER_RECURSIVE
    )

    def build_service(dependency: DecoratorDependency) -> DecoratorService:
        return DecoratorService(dependency=dependency)

    container.add_factory(
        build_service,
        provides=DecoratorService,
        dependency_registration_policy=DependencyRegistrationPolicy.REGISTER_RECURSIVE,
    )
    dependency_spec = container._providers_registrations.find_by_type(DecoratorDependency)
    assert dependency_spec is not None
    assert dependency_spec.concrete_type is DecoratorDependency


def test_resolve_autoregistration_integration_registers_dependency_chain() -> None:
    container = Container(
        missing_policy=MissingPolicy.REGISTER_RECURSIVE,
        dependency_registration_policy=DependencyRegistrationPolicy.REGISTER_RECURSIVE,
    )

    resolved = container.resolve(ResolveRoot)

    assert isinstance(resolved, ResolveRoot)
    assert isinstance(resolved.dependency, ResolveDependency)
    assert container._providers_registrations.find_by_type(ResolveDependency) is not None


def test_resolve_on_missing_false_keeps_missing_missing() -> None:
    container = Container(missing_policy=MissingPolicy.REGISTER_ROOT)

    with pytest.raises(DIWireDependencyNotRegisteredError):
        container.resolve(
            ResolveRoot,
            on_missing=MissingPolicy.ERROR,
        )

    assert container._providers_registrations.find_by_type(ResolveRoot) is None
    assert container._providers_registrations.find_by_type(ResolveDependency) is None


def test_resolve_on_missing_register_root_registers_only_root() -> None:
    container = Container()

    with pytest.raises(DIWireDependencyNotRegisteredError):
        container.resolve(
            ResolveRoot,
            on_missing=MissingPolicy.REGISTER_ROOT,
        )

    assert container._providers_registrations.find_by_type(ResolveRoot) is not None
    assert container._providers_registrations.find_by_type(ResolveDependency) is None


def test_resolve_on_missing_register_recursive_registers_chain() -> None:
    container = Container()

    resolved = container.resolve(
        ResolveRoot,
        on_missing=MissingPolicy.REGISTER_RECURSIVE,
    )

    assert isinstance(resolved, ResolveRoot)
    assert isinstance(resolved.dependency, ResolveDependency)
    assert container._providers_registrations.find_by_type(ResolveRoot) is not None
    assert container._providers_registrations.find_by_type(ResolveDependency) is not None


def test_resolve_from_container_uses_container_default_on_missing() -> None:
    container = Container(missing_policy=MissingPolicy.REGISTER_ROOT)

    with pytest.raises(DIWireDependencyNotRegisteredError):
        container.resolve(
            ResolveRoot,
            on_missing="from_container",
        )

    assert container._providers_registrations.find_by_type(ResolveRoot) is not None
    assert container._providers_registrations.find_by_type(ResolveDependency) is None


def test_resolve_provider_dependency_key_autoregisters_inner_dependency() -> None:
    container = Container(
        default_lifetime=Lifetime.TRANSIENT, missing_policy=MissingPolicy.REGISTER_ROOT
    )

    provider = container.resolve(Provider[_AutoProviderDependency])
    first = provider()
    second = provider()

    assert isinstance(first, _AutoProviderDependency)
    assert isinstance(second, _AutoProviderDependency)
    assert first is not second


@pytest.mark.asyncio
async def test_resolve_async_provider_dependency_key_autoregisters_inner_dependency() -> None:
    container = Container(
        default_lifetime=Lifetime.TRANSIENT, missing_policy=MissingPolicy.REGISTER_ROOT
    )

    provider = container.resolve(AsyncProvider[_AutoProviderDependency])
    first = await provider()
    second = await provider()

    assert isinstance(first, _AutoProviderDependency)
    assert isinstance(second, _AutoProviderDependency)
    assert first is not second


def test_resolve_provider_dependency_key_uses_existing_registration_without_recompile() -> None:
    container = Container(
        default_lifetime=Lifetime.TRANSIENT, missing_policy=MissingPolicy.REGISTER_ROOT
    )
    container.add(_AutoProviderDependency)
    container.compile()
    graph_revision_before = container._graph_revision

    provider = container.resolve(Provider[_AutoProviderDependency])
    resolved = provider()

    assert isinstance(resolved, _AutoProviderDependency)
    assert container._graph_revision == graph_revision_before


def test_resolve_provider_dependency_key_recompiles_when_inner_dependency_is_autoregistered() -> (
    None
):
    container = Container(
        default_lifetime=Lifetime.TRANSIENT, missing_policy=MissingPolicy.REGISTER_ROOT
    )
    previous_resolver = container.compile()

    provider = container.resolve(Provider[_AutoProviderDependency])
    resolved = provider()

    assert isinstance(resolved, _AutoProviderDependency)
    assert container._root_resolver is not None
    assert container._root_resolver is not previous_resolver


def test_resolve_reraises_when_missing_dependency_cannot_be_autoregistered() -> None:
    container = Container(missing_policy=MissingPolicy.REGISTER_ROOT)

    with pytest.raises(DIWireDependencyNotRegisteredError):
        container.resolve(str)


@pytest.mark.asyncio
async def test_aresolve_provider_dependency_key_autoregisters_then_reuses_compilation() -> None:
    container = Container(
        default_lifetime=Lifetime.TRANSIENT, missing_policy=MissingPolicy.REGISTER_ROOT
    )

    first_provider = await container.aresolve(Provider[_AutoProviderDependency])
    second_provider = await container.aresolve(Provider[_AutoProviderDependency])
    first = first_provider()
    second = second_provider()

    assert isinstance(first, _AutoProviderDependency)
    assert isinstance(second, _AutoProviderDependency)
    assert first is not second


@pytest.mark.asyncio
async def test_aresolve_autoregisters_on_miss_and_reraises_when_graph_unchanged() -> None:
    container = Container(
        missing_policy=MissingPolicy.REGISTER_RECURSIVE,
        dependency_registration_policy=DependencyRegistrationPolicy.REGISTER_RECURSIVE,
    )

    resolved = await container.aresolve(ResolveRoot)

    assert isinstance(resolved, ResolveRoot)
    with pytest.raises(DIWireDependencyNotRegisteredError):
        await container.aresolve(str)


def test_ensure_autoregistration_short_circuits_for_disabled_and_open_generic() -> None:
    strict_container = Container(
        missing_policy=MissingPolicy.ERROR,
        dependency_registration_policy=DependencyRegistrationPolicy.IGNORE,
    )

    strict_container._ensure_autoregistration(DirectDependency)

    assert strict_container._providers_registrations.find_by_type(DirectDependency) is None

    container = Container(missing_policy=MissingPolicy.REGISTER_ROOT)
    container.add_factory(_build_open_autoreg_dependency, provides=_OpenAutoregDependency)

    container._ensure_autoregistration(_OpenAutoregDependency[int])

    assert container._providers_registrations.find_by_type(_OpenAutoregDependency[int]) is None


def test_extract_provider_inner_dependency_fast_ignores_non_provider_metadata() -> None:
    container = Container()
    dependency = Annotated[DirectDependency, "plain-metadata"]

    assert container._extract_provider_inner_dependency_fast(dependency) is None


def test_resolve_autoregisters_settings_like_dependency_as_root_singleton(monkeypatch: Any) -> None:
    class _SettingsLikeDependency:
        value: str = "settings"

    monkeypatch.setattr(
        "diwire._internal.container.is_pydantic_settings_subclass",
        lambda candidate: candidate is _SettingsLikeDependency,
    )

    container = Container(missing_policy=MissingPolicy.REGISTER_ROOT)

    first = container.resolve(_SettingsLikeDependency)
    second = container.resolve(_SettingsLikeDependency)

    settings_spec = container._providers_registrations.get_by_type(_SettingsLikeDependency)
    assert first is second
    assert settings_spec.factory is not None
    assert settings_spec.concrete_type is None
    assert settings_spec.lifetime is Lifetime.SCOPED
    assert settings_spec.scope is Scope.APP


def test_resolve_autoregisters_pydantic_settings_as_singleton_factory() -> None:
    if pydantic_settings_base is None:
        pytest.skip("pydantic_settings is unavailable")

    container = Container(missing_policy=MissingPolicy.REGISTER_ROOT)

    first = container.resolve(PydanticSettingsDependency)
    second = container.resolve(PydanticSettingsDependency)

    settings_spec = container._providers_registrations.get_by_type(PydanticSettingsDependency)
    assert first is second
    assert settings_spec.factory is not None
    assert settings_spec.concrete_type is None
    assert settings_spec.lifetime is Lifetime.SCOPED
    assert settings_spec.scope is Scope.APP


def test_dependency_autoregistration_registers_pydantic_settings_as_root_singleton() -> None:
    if pydantic_settings_base is None:
        pytest.skip("pydantic_settings is unavailable")

    container = Container(
        dependency_registration_policy=DependencyRegistrationPolicy.REGISTER_RECURSIVE
    )

    container.add(
        RootWithPydanticSettingsDependency,
        scope=Scope.REQUEST,
        lifetime=Lifetime.SCOPED,
    )

    settings_spec = container._providers_registrations.get_by_type(PydanticSettingsDependency)
    assert settings_spec.factory is not None
    assert settings_spec.concrete_type is None
    assert settings_spec.lifetime is Lifetime.SCOPED
    assert settings_spec.scope is Scope.APP


def test_resolve_autoregisters_pydantic_v1_settings_as_singleton_factory() -> None:
    if pydantic_v1 is None:
        pytest.skip("pydantic.v1 is unavailable")
    pydantic_v1_module = cast("Any", pydantic_v1)
    base_settings_type = cast("type[Any]", pydantic_v1_module.BaseSettings)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=_PYDANTIC_V1_WARNING_PATTERN,
            category=UserWarning,
        )

        class PydanticV1SettingsDependency(base_settings_type):
            value: int = 7

        container = Container(missing_policy=MissingPolicy.REGISTER_ROOT)
        first = container.resolve(PydanticV1SettingsDependency)
        second = container.resolve(PydanticV1SettingsDependency)

    settings_spec = container._providers_registrations.get_by_type(PydanticV1SettingsDependency)
    assert first is second
    assert settings_spec.factory is not None
    assert settings_spec.concrete_type is None
    assert settings_spec.lifetime is Lifetime.SCOPED
    assert settings_spec.scope is Scope.APP


def test_concrete_registration_autoregisters_namedtuple_class_dependencies() -> None:
    class NamedTupleRoot(NamedTuple):
        dependency: NamedTupleFrameworkDependency

    container = Container()

    container.add(
        NamedTupleRoot,
        dependency_registration_policy=DependencyRegistrationPolicy.REGISTER_RECURSIVE,
    )

    dependency_spec = container._providers_registrations.find_by_type(
        NamedTupleFrameworkDependency,
    )
    assert dependency_spec is not None
    assert dependency_spec.concrete_type is NamedTupleFrameworkDependency


def test_concrete_registration_autoregisters_msgspec_struct_dependencies() -> None:
    class MsgspecRoot(msgspec.Struct):
        dependency: MsgspecFrameworkDependency

    container = Container()

    container.add(
        MsgspecRoot,
        dependency_registration_policy=DependencyRegistrationPolicy.REGISTER_RECURSIVE,
    )

    dependency_spec = container._providers_registrations.find_by_type(
        MsgspecFrameworkDependency,
    )
    assert dependency_spec is not None
    assert dependency_spec.concrete_type is MsgspecFrameworkDependency


def test_concrete_registration_autoregisters_pydantic_basemodel_v2_dependencies() -> None:
    pydantic_module = pytest.importorskip("pydantic")
    base_model_type = cast("type[Any]", pydantic_module.BaseModel)
    config_dict = cast("Any", pydantic_module.ConfigDict)

    class PydanticRoot(base_model_type):
        model_config = config_dict(arbitrary_types_allowed=True)
        dependency: PydanticFrameworkDependency

    container = Container()

    container.add(
        PydanticRoot,
        dependency_registration_policy=DependencyRegistrationPolicy.REGISTER_RECURSIVE,
    )

    dependency_spec = container._providers_registrations.find_by_type(
        PydanticFrameworkDependency,
    )
    assert dependency_spec is not None
    assert dependency_spec.concrete_type is PydanticFrameworkDependency
